/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

/* ****************************************************************************
   |      File: lbuffer.h
   |      Contains: class definition for LBuffer and LTemplBuffer
   |      Created: 20.03.1998
   |      Last Changed: 20.03.1998
   |      Version: 0.0
   |         
   |      ToDo: + just started
   |
   ************************************************************************* */

/* ====>> See also the ChangeLog in lsocket.cc <<==== */

#ifndef _L_BUFFER_H_
#define _L_BUFFER_H_

#include <lcl.h>
#include <lcl/lobject.h>

#include <stdio.h>

class LString;

/*
 * \class LBuffer
 * \author Jens Reimann
 * \brief class for managing void * buffers in C++ style
 *
 * This class encapsulates the well known void * buffer handling of C in a
 * (more or less good) C++ style.
 */

class LCL_EXPORT LBuffer : public LObject {

  //  friend class LString;

 protected:
  void * m_ptr;
  unsigned long m_size;

 public:
  LBuffer ();                              //!< Create a zero-size buffer
  LBuffer ( unsigned long size );          //!< Create an empty buffer of given size
  LBuffer ( unsigned long size, void * );  //!< Create a buffer of given size and fill with content
  LBuffer ( const LBuffer & );              //!< Copy constructor
  LBuffer ( const char * );                 //!< Create a buffer from a C string
  LBuffer ( LString const & );              //!< Create a buffer from a LString

  virtual ~LBuffer();                      //!< Destroy buffer and free memory

  const LBuffer & resize ( unsigned long size ); //!< resize the buffer
  const LBuffer & grow ( int );                  //!< grow the buffer

  operator void*();                         //!< return pointer to buffer
  operator char*();                         //!< return pointer to buffer
  char operator [](int i) const;            //!< access buffer a given position
  char operator [](unsigned long i) const;  //!< access buffer a given positionw
  char & operator [] (unsigned long i );    //!< access buffer a given positionw
  void * getPtr ( void ) const;             //!< return pointer to buffer

  void const * getConstPtr () const;
  unsigned long getSize () const; //!< get the current buffer size

  // ************ Operators *****************

  LBuffer & operator =(const char*);
  LBuffer & operator =(const LBuffer &);

  bool operator ==(const LBuffer & ) const;
  LBuffer operator +(const LBuffer &);
  void operator += (const LBuffer & );
  void operator += (char);

  void clear();                 //!< clear the buffer. Free memory. Set size to zero!
  void erase ();                //!< flood the buffer with zeros

  // ************ Append Operations ***********

  const LBuffer & append ( const LBuffer & );
  const LBuffer & append ( char const * );
  const LBuffer & append ( void const *, unsigned long );
  LBuffer & append ( char );
  LBuffer & append ( LString & );
  void prepend ( const LBuffer & );
  
  // ************ Cut Operations ************

  LBuffer cutFrom ( unsigned long from ) const;
  void cutFrom ( unsigned long from );

  // ************ File Operations *************

  virtual LBuffer & loadFile ( LString );   //!< Load buffer from file
  virtual LBuffer & loadFile ( NAMESPACE_STD(istream) & ); //!< Load buffer from file
  virtual LBuffer & loadFile ( FILE * );    //!< Load buffer from file
  void saveFile ( LString ) const;          //!< Dump buffer to file
  void saveFile ( NAMESPACE_STD(ostream) & ) const;        //!< Dump buffer to file
  void saveFile ( FILE * fp ) const;        //!< Dump buffer to file

  // ************ Type conversions *************

  LString toString () const;                //!< Convert buffer to a LString
};

NAMESPACE_STD(ostream) & operator <<( NAMESPACE_STD(ostream) & os, LBuffer p1 );


/* *****************************************************************************************
 * CLASS
 *              LMemberBuffer
 *
 * ***************************************************************************************** */

template <class T>
class LTemplBuffer : public LBuffer {

 public:

  LTemplBuffer ( T );
  LTemplBuffer ();
  LTemplBuffer ( void * );
  LTemplBuffer ( LBuffer & );

  virtual const LBuffer & resize ( unsigned long size );

  virtual LBuffer & loadFile ( LString str ) { LBuffer::loadFile(str);};
  virtual LBuffer & loadFile ( FILE * fp );
  virtual LBuffer & loadFile ( NAMESPACE_STD(istream) & ins );
  
  void set ( T );
  T * get () const;
};

template <class T>
LTemplBuffer<T>::LTemplBuffer ( T t )
{
  m_size = sizeof ( T );
  m_ptr = malloc ( size );

  set ( t );
}

template <class T>
LTemplBuffer<T>::LTemplBuffer ( void * _ptr )
{
  m_size = sizeof ( T );
  m_ptr = malloc ( m_size );
  memcpy ( m_ptr, _ptr, m_size );
}

template <class T>
LTemplBuffer<T>::LTemplBuffer ( )
{
  m_size = sizeof ( T );
  m_ptr = malloc ( m_size );
}

template <class T>
LTemplBuffer<T>::LTemplBuffer ( LBuffer & p1 )
{
  m_size = sizeof ( T );
  m_ptr = malloc ( m_size );

  memset ( m_ptr, 0, m_size );
  memcpy ( m_ptr, p1.getPtr(), MIN(m_size, p1.getSize() ) );
}

template <class T>
void LTemplBuffer<T>::set ( T t )
{
  memcpy ( m_ptr, &t, m_size );
}

template <class T>
T * LTemplBuffer<T>::get ( ) const
{
  return (T*)m_ptr;
}

template <class T>
const LBuffer & LTemplBuffer<T>::resize ( unsigned long _size )
{
  if (_size>m_size)
    LBuffer::resize(_size);

  return *this;
}

template <class T>
LBuffer & LTemplBuffer<T>::loadFile ( FILE * fp )
{

  if ( fp == NULL )
    return *this;  

  fread ( m_ptr, 1, m_size, fp );

  return *this;
}

template <class T>
LBuffer & LTemplBuffer<T>::loadFile ( NAMESPACE_STD(istream) & ins )
{
  char * buf = (char*)m_ptr;
  char c;
  int pos=0;

  while ( (c=ins.get())!=EOF && (pos<size) ) {
    buf[pos] = c;
    pos++;
  }
  
  return *this;

  /*  if (!ins) {

    clear();

  } else {

    ins.read ( ptr, size );

  }

  return *this;*/
}

template <class T>
inline NAMESPACE_STD(ostream) & operator <<( NAMESPACE_STD(ostream) & os, LTemplBuffer<T> p1 )
{
  os << *(p1.get());
  return os;
}

#include <lcl/lstring.h>

#endif
