#ifndef _LCOLLECTION_H_
#define _LCOLLECTION_H_

#include <lcl.h>
#include <lcl/lsimplelist.h>

class LCursor;

/*!
 * \defgroup Collections Collections
 */


/**
 * \class LBaseCollection
 * \brief Virtual base class for all collections
 * \author Jens Reimann <ctron@dentrassi.de>
 * \ingroup Collections
 *
 * It uses LSimpleList to handle internal data.
 *
 * For more information about handling data in lists (collections) see LCollection!
 *
 */

class LCL_EXPORT LBaseCollection {
  friend class LCL_EXPORT LCursor;
 protected:
  LSimpleList cursorList;

  virtual void addCursor ( LCursor * );                   //!< Adds a new cursor to the collection
  virtual void removeCursor ( LCursor * );                //!< Removes a cursor from the collection

 public:

  virtual ~LBaseCollection ();                            //!< Destructor of the class

  virtual void findFirst ( LCursor * ) const = 0;         //!< Sets the cursor to the first element
  virtual void findNext ( LCursor * ) const = 0;          //!< Sets the cursor to the next element
  virtual void findPrev ( LCursor * ) const = 0;          //!< Sets the cursor to the previous element
  virtual void findLast ( LCursor * ) const = 0;          //!< Sets the cursor to the last element
  virtual LCursor * createCursor ( ) const = 0;           //!< Crate a new cursor for this collection

  virtual void clear () = 0;                              //!< Clear the collection

  virtual void invalidate ();                             //!< Invalidates all cursors owned by this collection
  virtual void obsolete ();
};

#include <lcl/lcursor.h>

/**
 * \class LCollection
 * \brief Virtual base class for all collections
 * \author Jens Reimann <ctron@dentrassi.de>
 *
 * This is a base class for all templated lists in the LCL.
 *
 * Items may be added used the add member. Each list will handle an add request in a different way.
 *
 * For more information about the collections and how data is handled for each type of list see
 * LHeap, LSet, LMap.
 */

template <class Type>
class LCL_EXPORT LCollection : public LBaseCollection {
 protected:
  
 public:
  virtual bool add ( Type const & ) = 0;                  //!< Adds a new element to this collection

  virtual unsigned long entries () const = 0;             //!< Retrieve the number of elements in this collection

  virtual ~LCollection ();
};

template <class Type>
LCollection<Type>::~LCollection ()
{
}

#endif
