#ifndef _LDBSYSTEM_H_
#define _LDBSYSTEM_H_

#include <lcl/lpointer.h>
#include <lcl/lstring.h>
#include <lcl/lhash.h>
#include <lcl/lbaseobject.h>

/*!
 * \defgroup Database Database
 */

class LDBDriver;
class LDBQuery;

LCL_EXPORT LDBDriver * openDatabase ( LString const & );
//LDBDriver * openDatabase ( LConfigObject const & );

#include <lcl/ldbsystem.plugin.h>

/*!
 * \class LDBDriver
 * \ingroup Database
 * \brief Virtual base class of a database driver
 * \author Jens Reimann <ctron@dentrassi.de>
 *
 * This class is a base class for all database connections. Derived from this class are several database
 * drivers (MySQL, Oracle, ODBC). Each instance of this class will handle one database connection.
 *
 * The database driver will be compiled in the LCL based on the libraries you have on the build system.
 *
 * To create a new database connection use ::openDatabase
 */

class LCL_EXPORT LDBDriver : public LBaseObject {
 protected:
  static bool loaded_plugins;
  static LHash<LString,LDBDriver_PlugIn*> driverHash;

 public:
  static void loadPlugIns ();
  static LDBDriver_PlugIn * getDriver ( LString const & name );
  static void registerDriver ( LString const & name, LDBDriver_PlugIn * driver );

 protected:
  LHash<LString, LString> configHash;  

 public:

  virtual LManagedPointer<LDBQuery> createQuery () = 0;

  LDBDriver ( LHash<LString,LString> const & vars );
  virtual ~LDBDriver ();

  virtual LManagedPointer<LDBQuery> query ( LString const & sql );

  virtual bool connect () = 0;
  virtual bool disconnect () = 0;

  virtual bool connected () const = 0;

  LString getUser () const;
  LString getPassword () const;
  LString getDatabase () const;
  LString getDriverName() const;
  LString getOption (LString const &) const;

};

class LDBBind {
  protected:    
  public:
    virtual bool isValid() const = 0;

    virtual LString getValue () const = 0;
    virtual void setValue ( LString const & ) = 0;

};

/*!
 * \class LDBQuery
 * \ingroup Database
 * \brief Virtual base class of a database query
 */

class LCL_EXPORT LDBQuery : public LBaseObject {
 protected:
  LString m_sql;
  bool valid;

 public:
  LDBQuery ( LDBDriver & );
  LDBQuery ( LDBDriver &, LString const & );
  virtual ~LDBQuery ();

  virtual bool prepare ( LString const & );
  virtual bool exec ();
  virtual bool exec ( LString const & sql ) = 0;
  virtual bool exec ( LHash<LString, LString> const & );

  virtual long numFields () const = 0;

  virtual bool fetchNext ( LHash<LString, LString> & ) = 0;
  virtual bool fetchNext () = 0;

  virtual LString getField ( long ) const = 0;
  virtual LString getField ( LString const& ) const = 0;

  virtual LString getFieldName ( long ) const = 0;
  virtual long getFieldIdx ( LString const & ) const = 0;

  // liefert exec-wert zurck
  virtual bool isValid () const;

  virtual LManagedPointer < LDBBind > createBind ( LString const & name );
};

#endif

