/* ****************************************************************************
   |      File: lhash.h
   |      Contains: class definition for LHash
   |      Created: 08.09.1998
   |      Last Changed: 20.07.1999
   |      Version: 0.0
   |         
   |      ToDo: + just started
   |
   ************************************************************************* */

#ifndef _L_HASH_H_
#define _L_HASH_H_

#include <lcl.h>

template < class Key, class Element > class LCL_EXPORT LHash;

#include <lcl/ltable.h>
#include <lcl/lhashiter.h>
#include <lcl/lset.h>

template < class Key, class Element >
class LCL_EXPORT LHashContainer {
 public:
  Key key;
  Element element;

  LHashContainer ( Key const & k, Element const & e ) :
    key ( k ),
    element ( e )
    {}

  bool operator== ( LHashContainer<Key,Element> const & p1 ) const
    {
      return key == p1.key;
    }
  virtual ~LHashContainer () {};
};

template < class Key, class Element >
Key key ( LHashContainer<Key,Element> const & hc )
{
  return hc.key;
}

#include <lcl/lmap.h>

/*!
 * \class LHash
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A implementation of a hash table
 * \ingroup Collections
 *
 * LHash implements a simple hash table. The four main methods are set(), unset(), get() and isValid().
 *
 * Each class which shall be used as \b key for LHash need a seperate "hash function" in the form "unsigned long hash ( Value const &, unsigned long hash_size );"
 * (e.g. for "int" as key: unsigned long key ( int const & );) This "hash function" should return a "hash value" for this object.
 *
 * A "hash value" is used to locate an object stored in the hash table. It must not be a unique identifier since the hash table will compare the keys for
 * locating the element.
 *
 * A hash table has a specific hash size (n). Now the "hash value" does some kind of pre-selection into which area of the hash table the object will
 * be stored. Threfore the "hash function" must return a value between \b 0 and \b n).
 *
 * There is a default "hash function" which will take the binary content of the key and generate a hash value from it. In some cases this is not convinient.
 * For example the LString class stores a pointer to the actual string. It the default "hash function" would be used the pointer to the string would be used
 * to calculate the hash value. Since the value of the  pointer might change (but not the string) the item cannot be located later in the hash table.
 * 
 * Therefore the default "hash function" is specialized (unsigned long hash ( LString const &, unsigned long hash_size); ) to generate a hash value from
 * the actual characters of the string.
 *
 * There are default "hash functions" for all basic types (except pointers, char *, ...) and LString
 *
 * \note LHash support serialization using LSerialize
 */

template < class Key, class Element >
class LCL_EXPORT LHash /* : public LObject*/ {
 protected:
  LTable < LMap < LHashContainer<Key,Element>, Key > > hash_array; //!< The hash array
  unsigned long hash_size; //!< The size of the hash table
  unsigned long hash_entries; //!< The number of entries stored in the hash table

 public:
  
  LHash ( unsigned long _hash_size = 100 ); //!< The default constructor
  LHash ( LHash<Key,Element> const & ); //!< The copy constructor
 
  virtual ~LHash (); //!< The destructor

  virtual long addAllFrom ( LHash<Key, Element> & hash );   //!< Append all elements from another hash table
  virtual void allElementsDo ( LHashIterator<Key, Element> & ); //!< Iterator for all elements
  virtual void allElementsDo ( LConstHashIterator<Key, Element> & ) const; //!< Iterator for all elements (const)
  virtual unsigned long entries () const; //!< Fetch the number of entries stored in the hash table

  void clear (); //!< Clear the hash table

  bool isSet ( Key const & key )  const; //!< Check if a key is set in the hash table

  void unset ( Key const & key ); //!< Remove a key from the hash table
  bool set ( Key const & key, Element const & element ); //!< Set a key in the hash table and assign an element to it
  Element & get ( Key const & key ) const; //!< Get an element from the hash table by key
  Element const & get ( Key const & key, Element const & _default ) const; //!< Get an element from the hash table by key (const)

  LSet < Key > getKeys () const; //!< Get the list of keys stored in the hash table
};

#include <lcl/lxcpt.h>


template < class Key, class Element>
class AddHashIter : public LHashIterator<Key, Element> {
 public:
  LHash<Key, Element> & hash;
  unsigned long nums;
  AddHashIter ( LHash<Key, Element> & _hash ) :
    hash ( _hash ),
    nums ( 0 )
    {
    }

  void applyTo ( Key & key, Element & element )
    {
      hash.set ( key, element );
      nums++;
    }

};

template < class Key, class Value, class Serializer >
class FreezeHashIter : public LConstHashIterator < Key, Value >
{
 protected:  
 public:
  Serializer * ser;
  void applyTo ( Key const & key, Value const & value )
  {
    *ser << key;
    *ser << value;
  }
};

#include <lcl/lstring.h>

/*!
 * The default constructor of the hash table.
 *
 * The size of the hash table is a trade off between memory consumption and performance.
 * The default setting is approriate for most used where the hash table will not get much entries.
 * For hash tables which will receive a lot of entries (>5000) a increase of the hash size will consume more memory but
 * reduce the lookup speed of an element.
 *
 * Larger hash tables are faster when many entries are stored. Smaller hash tables will consume less memory for organization of the data.
 *
 * \param _hash_size the size of the hash table
 */
template < class Key, class Element >
LHash<Key, Element>::LHash ( unsigned long _hash_size ) :
  hash_array ( _hash_size ),
  hash_size ( _hash_size ),
  hash_entries ( 0 )
{
}

template < class Key, class Element >
LHash<Key, Element>::LHash ( LHash<Key,Element> const & p1 ) :
  hash_array ( p1.hash_array ),
  hash_size ( p1.hash_size ),
  hash_entries ( p1.hash_entries )
{
  
}

template < class Key, class Element >
LHash<Key, Element>::~LHash ()
{
}

/*!
 * \return the number of element stored in the hash table
 */

template < class Key, class Element >
unsigned long LHash<Key, Element>::entries () const
{
  return hash_entries;
}

/*!
 * \retval true the key is set in the hash table
 * \retval false the key is not set in the hash table
 */

template < class Key, class Element >
bool LHash<Key, Element>::isSet ( Key const & key )  const
{
  unsigned long hash_val = hash(key,hash_size);
  return hash_array [ hash_val ].containsElementWithKey ( key );
}

template <class Key, class Element>
long LHash<Key, Element>::addAllFrom ( LHash<Key, Element> & hash )
{
  AddHashIter<Key, Element> ahi ( *this );
  hash.allElementsDo ( ahi );
  return ahi.nums;
}

/*!
 * \post the hash table will be empty and contains zero elements
 */
template < class Key, class Element >
void LHash<Key, Element>::clear ()
{
  luint32 i;
  for ( i = 0; i<hash_size; i++) {
    hash_array [i].clear ();
  }
  hash_entries = 0;
}

/*!
 * Fetches an element from the hash table by key.
 *
 * If the key is not set the method wil throw a LBasicException.
 *
 * \param key the key which is used to look up the element
 *
 * \return the element which is assigned to the provided key
 *
 */
template <class Key, class Element >
Element & LHash<Key,Element>::get ( Key const & key ) const
{
  unsigned long hash_val = hash ( key, hash_size );
  
  if ( !hash_array [ hash_val ].containsElementWithKey ( key ) ) {
    //    throw "Element & LHash<>::get ( Key const &) const";
    throw LBasicException ( __FILE__, __LINE__, "Key not found" );
    MSTHROW(Element);
  }

  return hash_array [ hash_val ].elementWithKey ( key ).element;
}

/*!
 * Fetches an element from the hash table by key.
 *
 * If the key is not set the method will return the  provided default value.
 *
 * \param key the key which is used to look up the element
 * \param _default the default value which will be returned if the key is not set 
 * \return the element which is assigned to the provided key or the provided default value if the key is not set
 *
 * \code

LHash<int,int> hash;

hash.set ( 1, 123 );
hash.set ( 2, 456 );

int i;

i = hash.get ( 1, 0 ); // i = 123
i = hash.get ( 3, 0 ); // i = 0 (default value)

\endcode
 */
template <class Key, class Element >
Element const & LHash<Key,Element>::get ( Key const & key, Element const & _default ) const
{
  if ( isSet ( key ) ) {
    return get ( key );
  } else {
    return _default;
  }
}

/*!
 * Remove an element from the hash table by key
 *
 * If the key does not exists, nothing will happen.
 *
 * \param key the key which should be removed
 *
 * \post The element assigned to this key will be removed. The is no defined anymore.
 */

template < class Key, class Element >
void LHash<Key, Element>::unset ( Key const & key )
{
  unsigned long hash_val = hash(key,hash_size);
  if ( hash_array [ hash_val ].containsElementWithKey ( key ) ) {
    hash_entries--;
    hash_array [ hash_val ].removeElementWithKey ( key );
  }
}

/*!
 * Sets an element to a key. If the key is already set with another element it will be overwritten.
 *
 * \param key the key to set
 * \param element the element to set
 *
 * \post The key will have the element provided
 *
 * \code

LHash<int, LString> hash;

hash.set ( 1, "FOO" ); // 1 is now set to "FOO"
hash.set ( 2, "BAR" ); // 2 is now set to "BAR"
hash.set ( 1, "BAR" ); // 1 is now set to "BAR"

\endcode
 *
 */

template <class Key, class Element>
bool LHash <Key, Element >::set ( Key const & key, Element const & element )
{
  //  unsigned long hash_val = hash((const char*)&key,sizeof(key),hash_size);
  unsigned long hash_val = hash ( key, hash_size );

  try {
    
    // try to replace ...
    if (!hash_array [ hash_val ].replaceElementWithKey ( key, LHashContainer<Key,Element>(key, element ))) {
      
      // ... otherwise insert!
      if ( hash_array [ hash_val ].add( LHashContainer<Key,Element>(key, element) ) ) {
	hash_entries++;
	return true;
      } else { // if insert
	return false;
      }
      
    } else { // if replace
      
      return true;
      
    }
    
  }
  
  // oops!
  catch (...) {
    return false;
  }      
}

template <class Key, class Element>
void LHash<Key, Element>::allElementsDo ( LHashIterator<Key, Element> & iter )
{
  unsigned long i;
  LCursor * cursor;
  for ( i=0; i<hash_size; i++) {
    cursor = hash_array[i].createCursor ();
    for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
      iter.applyTo ( hash_array[i].elementAt(cursor).key, hash_array[i].elementAt(cursor).element );
    }
    delete cursor;
  }
}

template <class Key, class Element>
void LHash<Key, Element>::allElementsDo ( LConstHashIterator<Key, Element> & iter ) const
{
  unsigned long i;
  LCursor * cursor;
  for ( i=0; i<hash_size; i++) {
    cursor = hash_array[i].createCursor ();
    for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
      iter.applyTo ( hash_array[i].elementAt(cursor).key, hash_array[i].elementAt(cursor).element );
    }
    delete cursor;
  }
}

template < class Key, class Element> 
class GetHashKeysIter : public LConstHashIterator<Key,Element>
{
 protected:
 public:
  LSet < Key > keys;
  void applyTo ( Key const & key, Element const & item )
    {
      keys.add ( key );
    }
  virtual ~GetHashKeysIter () {};
};

/*!
 *
 * Retrieve all keys from the hash table. Since keys are defined to be unique the will be returned
 * in a LSet<Key> object.
 * 
 * \return a list of all keys in the hash table
 */

template < class Key, class Element >
LSet < Key > LHash<Key,Element>::getKeys () const
{
  GetHashKeysIter<Key,Element> iter;
  allElementsDo ( iter );
  return iter.keys;
}

template <class Key>
unsigned long hash ( Key const & key, unsigned long hash_size )
{
  return hash ( (char const *)&key, sizeof ( key ), hash_size );
}


template < class Key, class Value, class Serializer >
Serializer & operator << ( Serializer & ser, LHash<Key,Value> const & hash );

template < class Key, class Value, class Serializer >
Serializer & operator << ( Serializer & ser, LHash<Key,Value> const & hash )
{
  Serializer _ser;
  FreezeHashIter<Key,Value,Serializer> fhi;
  fhi.ser = &_ser;
  hash.allElementsDo ( fhi );
  ser << *fhi.ser;

  return ser;
}

template < class Key, class Value, class Serializer >
Serializer & operator >> ( Serializer & ser, LHash<Key,Value> & hash )
{

  hash.clear ();

  Serializer _ser;
  ser >> _ser;

  try
    {
      while ( true )
	{
	  Key key;
	  Value value;
	  _ser >> key;
	  _ser >> value;
	  hash.set ( key, value );
	}
    }
  catch ( ... )
    {
    }
  return ser;
}


#endif
