/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

#ifndef _LHEAP_H_
#define _LHEAP_H_

#include <lcl/lcollection.h>
#include <lcl/llinkedcont.h>
#include <lcl/llinkedseq.h>

#include <lcl/literator.h>

#include <lcl/lcursor.h>

/**
 * \class LHeap
 * \brief A collection class which contains multiple, unsorted elements.
 * \author Jens Reimann <ctron@dentrassi.de>
 * \version 0.2
 * \date 1999
 * \bug Should be very bug free.
 * \ingroup Collections
 *
 * This is a unsorted collection which supports multiple elements. The elements used with this collection don't need to
 * have the operator == defined. On the other side it is not possible to retreive an object directly from the collection. But
 * it is possible to iterate through the collection and process all entries.
 *
 * A LLinkedSequence is use as storage.
 */

template <class Type>
class LCL_EXPORT LHeap : public LCollection<Type> {
 protected:
  LLinkedSequence<Type, LLinkedContainer<Type> > sequence;        //!< The storage for this class

  LLinkedCursor<Type> * baseCursor ( LCursor * ) const;           //!< Converts the given base cursor to a cursor of this class

 public:
  LHeap ();                                                       //!< Constructor
  LHeap ( LHeap<Type> const & );                                  //!< Copy Constructor
  virtual ~LHeap();                                               //!< Destructor

  virtual LCursor * createCursor () const;                        //!< Creates a new cursor which is owned by this collection

  virtual bool add ( Type const & );                              //!< Add a new element to this collection
  virtual bool prepend ( Type const & );                          //!< Add a new element to the head of the collection
  virtual unsigned long entries () const;                         //!< Retrieves the number of elements in this collection

  virtual void findFirst ( LCursor * ) const;                     //!< Sets the cursor to the first element
  virtual void findNext ( LCursor * ) const;                      //!< Sets the cursor to the next element
  virtual void findPrev ( LCursor * ) const;                      //!< Sets the cursor to the previous element
  virtual void findLast ( LCursor * ) const;                      //!< Sets the cursor to the last element

  virtual Type & elementAt ( LCursor * );                         //!< Retrieves the element at this cursor position
  virtual Type const & elementAt ( LCursor * ) const;             //!< Retrieves the element at this cursor position (as const)

  virtual bool removeElementAt ( LCursor * );                     //!< Removes the element at the cursor position from the collection

  virtual void allElementsDo ( LIterator<Type> & );               //!< Iterates through all elements

  virtual void clear ();                                          //!< Clear this collection

  void dump () const;
};

template <class Type>
LHeap<Type>::LHeap ()
{  
}

template <class Type>
LHeap<Type>::LHeap ( LHeap const & p1 ) :
  sequence ( p1.sequence )
{  
}

template <class Type>
LHeap<Type>::~LHeap()
{
}

template <class Type>
bool LHeap<Type>::add ( Type const & item )
{
  LLinkedContainer<Type> * container = new LLinkedContainer<Type>( item );
  sequence.addContainer ( container );
  invalidate ();
  return true;
}

template <class Type>
bool LHeap<Type>::prepend ( Type const & item )
{
  LLinkedContainer<Type> * container = new LLinkedContainer<Type> ( item );
  sequence.prependContainer ( container );
  invalidate ();
  return true;
}

template <class Type>
unsigned long LHeap<Type>::entries () const
{
  return sequence.items;
}

template <class Type>
void LHeap<Type>::clear ()
{
  sequence.clear ();
  invalidate ();
}

template <class Type>
LLinkedCursor<Type> * LHeap<Type>::baseCursor ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) )
    return (LLinkedCursor<Type>*)cursor;
  return NULL;
}

template <class Type>
void LHeap<Type>::dump ( ) const
{
  LLinkedContainer<Type> * current = sequence.head;

  while ( current ) {
    cout << current->item << "\n";
    current = current->next;
  }
}

template <class Type>
void LHeap<Type>::findFirst ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = sequence.head;
    baseCursor(cursor)->valid = true;
  }
}

template <class Type>
void LHeap<Type>::findNext ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    baseCursor(cursor)->container = baseCursor(cursor)->container->next;
  }
}

template <class Type>
void LHeap<Type>::findLast ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) ) {
    baseCursor(cursor)->container = sequence.tail;
    baseCursor(cursor)->valid = true;
  }
}

template <class Type>
void LHeap<Type>::findPrev ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    baseCursor(cursor)->container = baseCursor(cursor)->container->prev;
  }
}

template <class Type>
LCursor * LHeap<Type>::createCursor () const
{
  return new LLinkedCursor<Type>( (LBaseCollection*)this);
}

template <class Type>
Type & LHeap<Type>::elementAt ( LCursor * cursor )
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return baseCursor ( cursor )->container->item;
  }
  throw 1;
  MSTHROW(Type);
}

template <class Type>
Type const & LHeap<Type>::elementAt ( LCursor * cursor ) const
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    return baseCursor ( cursor )->container->item;
  }
  throw 1;
  MSTHROW(Type);
}

template <class Type>
void LHeap<Type>::allElementsDo ( LIterator<Type> & p_iterator )
{
  LLinkedContainer<Type> * current = sequence.head;
  while ( current ) {
    p_iterator.applyTo ( current->item );
    current = current->next;
  }
}

template <class Type>
inline bool LHeap<Type>::removeElementAt ( LCursor * cursor )
{
  if ( cursor->isFor ( this ) && cursor->isValid() ) {
    sequence.removeContainer ( baseCursor(cursor)->container );
    delete baseCursor(cursor)->container;
    invalidate ();
    return true;
  } else {
    return false;
  }
}


#endif
