/* -*-mode: c++; coding: iso-8859-2-unix; -*- */

/* LCL - The Linux Class Library
 * Copyright (C) 1998-2003 Jens Reimann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef _LLINKEDSEQUENCE_H_
#define _LLINKEDSEQUENCE_H_

#include <lcl.h>
#include <lcl/lxcpt.h>

/*!
 * \class LLinkedSequence
 * \ingroup Collections
 * \author Jens Reimann <ctron@dentrassi.de>b
 *
 * This is a very generic implementation of a double linked list. It is mostly used by the
 * collections (LHeap, LSet, LMap, LHash). A linked sequence only knows about containers and not
 * about the actual data which should be stored in a sequence. It is designed as a helper class for
 * high level collections.
 *
 * If you want to store data in collections use LHeap, LSet, LMap or LHash instead!
 *
 * The linked list stores data items in containers. The container has to be specified when instantiating
 * the class. In most cases LLinkedContainer will be used.
 *
 * A container needs a next and prev pointer which will point to the next or previous item in the linked list.
 * If either next or prev is NULL the container is head (or tail) of the sequence.
 *
 * It also will define (otherwise it would be very stupid) a data member (which will not be used by
 * LLinkedSequence) which will reflect the data wrapped in the container.
 *
 */

template <class Type, class Container>
class LCL_EXPORT LLinkedSequence {
 public:
  Container *head, *tail;
  unsigned long items;

 public:
  LLinkedSequence ();
  LLinkedSequence ( LLinkedSequence<Type,Container> const & );
  virtual ~LLinkedSequence ();

  virtual void addContainer ( Container * container );              //!< Add a new container to the sequence
  virtual void prependContainer ( Container * container );          //!< Add a new container to the head of the sequence

  virtual void addContainerBefore ( Container * add, Container * pos );
  virtual void addContainerAfter ( Container * add, Container * pos );

  virtual void removeContainer ( Container * );           //!< Remove a container which is in the sequence
  virtual void removeContainerAndDelete ( Container * );  //!< Remove a container which is in the sequence and delete it afterwards

  //! The assignment operator of the sequence
  LLinkedSequence<Type, Container> const & operator = ( LLinkedSequence<Type, Container> const & );

  //! Clears the sequence and frees all containers
  virtual void clear ();
};

template <class Type, class Container>
LLinkedSequence<Type, Container>::LLinkedSequence () :
  head ( NULL ), tail ( NULL ), items ( 0L )
{
}

template <class Type, class Container>
LLinkedSequence<Type,Container>::LLinkedSequence ( LLinkedSequence<Type,Container> const & p1 ) :
  head ( NULL ), tail ( NULL ), items ( 0L )
{
  Container * current = p1.head;
  while ( current ) {
    addContainer ( new Container ( *current ) );
    current = current->next;
  }
}

template <class Type, class Container>
LLinkedSequence<Type, Container>::~LLinkedSequence ()
{
  clear ();
}

template <class Type, class Container>
LLinkedSequence<Type, Container> const & LLinkedSequence<Type,Container>::operator = ( LLinkedSequence<Type, Container> const & p1 )
{
  clear ();

  Container * current = p1.head;
  while ( current ) {
    addContainer ( new Container ( *current ) );
    current = current->next;
  }

  return *this;
}

/*!
 * \param container The container which should be appended
 *
 * It will append a new container to the sequence. The container will then be maintained by the sequence.
 * The object may not be deleted while it is in the sequence!
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::addContainer ( Container * container )
{
  if ( container == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );
  
  if ( head == NULL ) {
    head = tail = container;
  } else {

    lcl_assert ( tail != NULL );

    container->prev = tail;
    container->next = NULL;
    tail->next = container;    
    tail = container;

  }

  items++;
}

template <class Type, class Container>
void LLinkedSequence<Type, Container>::prependContainer ( Container * container )
{
  if ( container == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );

  if ( head == NULL )
    {
      head = tail = container;
    }
  else
    {
      lcl_assert ( tail != NULL );

      container->prev = NULL;
      container->next = head;
      head->prev = container;
      head = container;
    }

  items++;
}

template <class Type, class Container>
void LLinkedSequence<Type, Container>::addContainerBefore ( Container * add, Container * pos )
{
  if ( add == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );
  if ( pos == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );

  if ( pos->prev )
    {
      add->prev = pos->prev;
      pos->prev->next = add;
      add->next = pos;
      pos->prev = add;
    }
  else
    {
      // Pos is the head item!
      prependContainer ( add );
    }

  items++;
}

template <class Type, class Container>
void LLinkedSequence<Type, Container>::addContainerAfter ( Container * add, Container * pos )
{
  if ( add == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );
  if ( pos == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );

  if ( pos->next )
    {
      add->prev = pos;
      pos->next->prev = add;
      add->next = pos->next;
      pos->next = add;
    }
  else
    {
      // Pos is the head item!
      addContainer ( add );
    }


  items++;
}


/*!
 * All entries which a currently in the sequence will be removed. All containers will be deleted and the
 * memory will be freed.
 *
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::clear ( )
{
  Container *current = head, *temp;
  while ( current ) {
    temp = current->next;
    delete current;
    current = temp;
  }

  head = tail = NULL;
  items = 0;
}


// ***********************************************************************************************
// **  Delete container from linked list
// ***********************************************************************************************

/*!
 * \param container The container which should be removed from the sequence
 *
 * The specified container will be removed from the sequence. The container will not be deleted and
 * the memory is not freed! You have do to this manually using delete!
 *
 * If you want to free the container automatically you need to call LLinkedSequence::removeContainerAndDelete
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::removeContainer ( Container * container )
{
  if ( container == NULL )
    throw LNullPointerException ( __FILE__, __LINE__ );

  Container * current = head;

  while ( current ) {
    if ( current == container ) {

      if ( current->prev ) {
	current->prev->next = current->next;
      } else {
	head = current->next;
      }
      if ( current->next ) {
	current->next->prev = current->prev;       
      } else {
	tail = current->prev;
      }
      
    }
    current = current->next;
  }

  items--;
}

/*!
 * \param container The container which should be removed from the sequence
 *
 * The specified container will be removed from the sequence. The container will be deleted and the memory
 * freed.
 */

template <class Type, class Container>
void LLinkedSequence<Type, Container>::removeContainerAndDelete ( Container * container )
{
  removeContainer ( container );
  delete container;
}

#endif

