/* ****************************************************************************
   |      File: lmutex.h
   |      Contains: class definition for LMutex
   |      Created: 30.06.1998
   |      Last Changed: 30.06.1998
   |      Version: 0.0
   |         
   |      ToDo: + not very much
   |
   ************************************************************************* */

/* ====>> See also the ChangeLog in lmutex.cc <<==== */

#ifndef _L_MUTEX_H_
#define _L_MUTEX_H_

#include <lcl.h>
#include <lcl/lobject.h>
#include <lcl/lvcs.h>

#ifdef _LCL_SYS_WIN32_
#include <windows.h>
#else
#include <pthread.h>
#endif

class LCL_EXPORT LMutex;

/*!
 * \defgroup IPC IPC
 */

/*!
 * \ingroup IPC
 * \class LMutex
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A mutex lock
 *
 * A simple mutex lock wrapper.
 *
 * A mutex lock may be locked. If it is already locked the call to lock will be blocked
 * until the thread which called lock previously calls unlock.
 */

class LCL_EXPORT LMutex : public LObject {
  friend class LCondVar;
  LVCS_DECLARE;
 protected:

  #ifdef _LCL_SYS_WIN32_
  HANDLE mutxp;
  #else
  pthread_mutex_t mutxp;
  #endif

 public:
  LMutex ();      //!< Creates a new mutex lock
  ~LMutex ();     //!< Destroys the mutex lock

  void lock () const;   //!< Lock the mutex
  bool trylock () const; //!< Try to lock the mutex
  void unlock () const; //!< Unlock the mutex

};

#ifndef NEED_FALLBACK_GUARD_MUTEX
class LCL_EXPORT LGuardMutex : public LMutex {
 private:
  #ifdef _LCL_SYS_POSIX_
  static pthread_mutexattr_t guard_attr;
  static bool guard_attr_init;
  static void checkInit ();
  #endif

  #ifdef _LCL_SYS_WIN32_
  CRITICAL_SECTION crit;
  #endif

 public:
  LGuardMutex ();

  #ifdef _LCL_SYS_WIN32_  
  virtual ~LGuardMutex ();
  virtual void lock () const;
  virtual void unlock () const;
  virtual bool trylock () const;
  #endif
};
#else

class LCL_EXPORT __lcl_fallbackGuardMutex;

class LCL_EXPORT LGuardMutex : public LMutex {
 private:
  __lcl_fallbackGuardMutex * mtx;
 public:
  LGuardMutex ();
  ~LGuardMutex ();
  virtual void lock () const;
  virtual void unlock () const;
  virtual bool trylock () const;
};
#endif

/*!
 * \ingroup IPC
 * \class LAutoLock
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A auto lock based on LMutex
 *
 * This class provides an interface to a LMutex. It locks the mutex in contruction and
 * frees is on destruction. It may be used to lock a mutex simply by creating an instance
 * and will unlock the mutex at end of lifetime.
 *
 */

class LCL_EXPORT LAutoLock : public LObject {
  LVCS_DECLARE;

 protected:
  LMutex const & mutex;

  bool m_attached;

 public:
  LAutoLock ( LMutex const & _mutex );
  ~LAutoLock ();

  void detach ();
};

//#include <lcond.h>

#endif



