#ifndef _LPLUGIN_H_
#define _LPLUGIN_H_

#include <lcl/llibrary.h>
#include <lcl/lstring.h>
#include <lcl/ltrace.h>
#include <lcl/lhash.h>
#include <lcl/lvcs.h>

/*!
 * \defgroup PlugIns Plug-Ins
 */

/*!
 * \struct _lcl_plugin_info
 *
 * This structure contains all plug-in information
 */
typedef struct _lcl_plugin_info {
  char module[128];
  char description[4*1024];
  unsigned long group;
} lcl_plugin_info;

/*!
 * \class LGenericPlugIn
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A Generic PlugIn Loader
 * \ingroup PlugIns
 *
 * This class loads a shared library using LLibrary and reads out the
 * LCL PlugIn information
 */
class LCL_EXPORT LGenericPlugIn {
  LVCS_DECLARE;
 protected:
  LLibrary lib;
  lcl_plugin_info info;

  void loadInfo ( void );

 public:
  LGenericPlugIn ( LString const & fname );

  virtual bool isValid() const;              //!< Check if the plug in is valid

  LString getName () const;                  //!< Return the name of the plug in
  LString getDescription () const;           //!< Return the description of the plug in
  LString getGroup () const;                 //!< Return the group of the plug in
};

/*!
 * \class LPlugIn
 * \author Jens Reimann <ctron@dentrassi.de>
 * \brief A specialized plug in class
 * \ingroup PlugIns
 *
 * This class loads a plug in based of LGenericPlugIn and does some
 * type safety checks if this plug in belongs to the correct group.
 *
 * It also provides an interface to create new objects from the plug in.
 *
 * In the plug in you should use the macros LCL_PLUGIN_INIT and LCL_PLUGIN_FACTORY,
 * LCL_PLUGIN_FACTORY_WITH_ARGS to create the common plug in functions.
 *
 * To create a new plug in for select a base class for all different objects of this
 * plug in (e.g. LDBDriver). Next create different source files (one or more for each
 * plug in). Derieve the specialized class from your base class ( e.g. LDBDriver_ODBC from
 * LDBDriver). Implement this object in the plug in source files.
 *
 * In the plug in source execute the macro LCL_PLUGIN_INIT once, setting the plug in information.
 * And LCL_PLUGIN_FACTORY and/or LCL_PLUGIN_FACTORY_WITH_ARGS to create the code for
 * creating a new instance of the class.
 *
 * When your create a new plug in using LPlugIn::LPlugIn the plug in information is read out of
 * the library. If the group ID matches the plug in is valid.
 *
 * When you call LPlugIn::createObject it will call lcl_plugin_get_factory which will create
 * a new instance (from the derieved class) inside the plug in.
 *
 * Example: The database system first loads all plug ins with a matching group ID (2). And
 * inserts the plug ins into a hash. When a database driver is needed it looks into the hash
 * to see if the driver has been loaded. If the driver is loaded it will call
 * LDBDriver_PlugIn::createObject which will create a new instance (e.g. LDBDriver_ODBC) inside
 * the plug in.
 *
 * You may then use the LDBDriver instance (which is acutally LDBDriver_ODBC then).
 */

template < class T, unsigned long group >
class LCL_EXPORT LPlugIn : public LGenericPlugIn {
 protected:
 public:

  LPlugIn ( LString const & fname );

  virtual bool isValid() const;

  virtual T * createObject ();
  virtual T * createObject ( LHash<LString,LString> const & args );
};

template < class T, unsigned long group >
LPlugIn<T,group>::LPlugIn ( LString const & fname ) :
  LGenericPlugIn ( fname )
{
}

/*!
 * \return true if the library is OK
 *
 * Check if the library is valid. Valid means the plug in and library are loaded successfully
 * and the group ID matches to the requested one.
 */
template < class T, unsigned long group >
bool LPlugIn<T,group>::isValid() const
{
  return LGenericPlugIn::isValid() && (info.group == group);
}

/*!
 * \return the pointer to the newly created instance
 *
 * Creates a new object inside the plug in. This object is derieved from the base class T (template
 * argument)
 */
template < class T, unsigned long group >
T * LPlugIn<T,group>::createObject ()
{
  if ( !isValid() )
    {
      LTRACE ( "LPlugIn", "PlugIn not loaded or not compatible!" );
      return NULL;
    }

  void * (*factory_proc)(void);

  factory_proc = (void*(*)(void))lib.getProc ( "lcl_plugin_factory" );

  if ( factory_proc )
    {
      LTRACE ( "LPlugIn", "Creating new object" );
      return (T*)factory_proc ();
    }
  else
    {
      LTRACE ( "LPlugIn", "Factory proc not found!" );
      return NULL;
    }
}

/*!
 * \return the pointer to the newly created instance
 *
 * Creates a new object inside the plug in. This object is derieved from the base class T (template
 * argument).
 *
 * This version of createObject will pass a LString hash to the plug in which may be used
 * for creation of the new instance. (e.g. LDBDriver_PlugIn sends to login information
 * to the plug in using this method)
 */
template < class T, unsigned long group >
T * LPlugIn<T,group>::createObject ( LHash<LString,LString> const & args )
{
  if ( !isValid() )
    {
      LTRACE ( "LPlugIn", "PlugIn not loaded or not compatible!" );
      return NULL;
    }

  void * (*factory_proc)(void*);

  factory_proc = (void*(*)(void*))lib.getProc ( "lcl_plugin_factory_with_args" );

  if ( factory_proc )
    {
      LHash<LString,LString> hash ( args );
      LTRACE ( "LPlugIn", "Creating new object" );
      return (T*)factory_proc ( (void*)&hash );
    }
  else
    {
      LTRACE ( "LPlugIn", "Factory proc not found!" );
      return NULL;
    }
}

#ifdef _WIN32
#define LCL_PLUGIN_EXPORT __declspec(dllexport)
#else
#define LCL_PLUGIN_EXPORT 
#endif

/*!
 * \param n the name of the plug in (e.g. "ODBC" for ODBC database driver)
 * \param d the description of the plug in (any descriptive text)
 * \param g the group ID of the plug in (e.g. 2 for database drivers)
 *
 * Create the function which returns the plug in information
 *
 * The following groups are defined: (1) for test (2) for database (3) for images (4) trace modules
 * (5) for config modules (6) for parsers (7) for interpreter commands
 *
 * For user defined groups start at 100000!
 */
#define LCL_PLUGIN_INIT(n,d,g) static lcl_plugin_info info; extern "C" LCL_PLUGIN_EXPORT lcl_plugin_info * lcl_plugin_get_info ( void ) { strncpy ( info.module, n, 128 ); strncpy ( info.description, d, 4*1024 ); /*strncpy ( info.group, g, 128 )*/ info.group=g; return &info; }

/*!
 * \param t the class of the plug in object
 *
 * creates the function code to create a new class using the default constructor
 */
#define LCL_PLUGIN_FACTORY(t) extern "C" LCL_PLUGIN_EXPORT void * lcl_plugin_factory ( void ) { return new t; }

/*!
 * \param t the class of the plug in object
 *
 * creates the function code to create a new class using the constructor Object ( LHash<LString,LString> const & )
 */

#define LCL_PLUGIN_FACTORY_WITH_ARGS(t) extern "C" LCL_PLUGIN_EXPORT void * lcl_plugin_factory_with_args ( void * args ) { return new t ( *(LHash<LString,LString>*)args); }


#endif
