#ifndef _LTRACE_CHANNEL_H_
#define _LTRACE_CHANNEL_H_

class LTraceChannel;

#include <lcl.h>

#include <lcl/lpointer.h>
#include <lcl/lhash.h>

/*!
 * \class LTraceChannel
 * \author Jens Reimann
 * \ingroup Tracing
 * \brief Base class for all tracing modules
 *
 * A base class for all other tracing modules.
 *
 * New trace modules must be added using LTraceChannelManager::addChannel
 */

class LTraceChannel {
 protected:
  bool active;

 public:

  LTraceChannel ();
  virtual ~LTraceChannel ();

  void enable ( bool = true );
  void disable ();

  bool isActive () const;

  virtual void trace ( const char * module, const char * message ) = 0;

};

/*!
 * \class LTraceChannel_OStream
 * \author Jens Reimann
 * \ingroup Tracing
 * \brief Tracing module which sends to the output formatted to a LOStream opened by this module (e.g. a file or the console)
 */

class LTraceChannel_OStream : public LTraceChannel {
 protected:
  LOStream * ostream;
  bool valid;
  
 public:
  LTraceChannel_OStream ( LHash<LString,LString> & );
  virtual void trace ( const char * module, const char * message );
};

/*!
 * \class LTraceChannelManager
 * \author Jens Reimann
 * \ingroup Tracing
 * \brief The manager of all trace channels
 *
 * This class manages all trace channels in an application. There may be only one instance per application
 * (singleton pattern). Trace channels may be added/removed during runtime.
 *
 */

class LTraceChannelManager {
 protected:
  LHash<LString, LManagedPointer<LTraceChannel> > channelHash;

  static LTraceChannelManager * theInstance;

  LTraceChannelManager ();

 public:
  static LTraceChannelManager * getInstance ();

  void trace ( const char * module, const char * message );

  bool addChannel ( LString const & channelName, LString const & channel );
  bool removeChannel ( LString const & channelName );

  LManagedPointer<LTraceChannel> getDefaultTraceChannel () const;
};

#endif
