#include <lcl/lasciiproto.h>

// --------->>>> LASCIIProtocolHandler

LASCIIProtocolHandler::LASCIIProtocolHandler () :
  m_connection ( NULL )
{
}

LASCIIProtocolHandler::~LASCIIProtocolHandler ()
{
}

void LASCIIProtocolHandler::setConnection ( LASCIIProtocolConnection * connection )
{
  m_connection = connection;
}

// --------->>>> LASCIIProtocolFactory

LASCIIProtocolFactory::LASCIIProtocolFactory ()
{
}

LASCIIProtocolFactory::~LASCIIProtocolFactory ()
{
}

// --------->>>> LASCIIProtocolFactory

LASCIIProtocolConnection::LASCIIProtocolConnection ( LASCIIProtocolHandler * handler, LFileDesc const & fileDesc ) :
  m_handler ( handler ),
  m_fileDesc ( fileDesc ),
  m_connected ( fileDesc.isValid() ),
  m_closeMarker ( false ),
  m_timeout ( 120 ),
  m_lastData ( time(0) ),
  m_skipLine ( false )
{

  m_handler->setConnection ( this );

  m_readSlot = connect_emitter_1 ( &LASCIIProtocolConnection::eventRead, *this, m_fileDesc.readEvent );
  m_writeSlot = connect_emitter_1 ( &LASCIIProtocolConnection::eventWrite, *this, m_fileDesc.writeEvent );
  m_timeoutSlot = connect_emitter_0 ( &LASCIIProtocolConnection::eventSelectTimeout, *this, LSelectLoop::getInstance()->timeoutEvent );

  m_fileDesc.reg ( LSelectLoop::getInstance() );

  checkEventMasks ();

}

LASCIIProtocolConnection::~LASCIIProtocolConnection ()
{

  close ();

  if ( m_handler )
    {
      delete m_handler;
      m_handler = NULL;
    }
}

void LASCIIProtocolConnection::sendLine ( LString const & line )
{
  LAutoLock _lock ( m_sendLock );

  if ( !m_closeMarker )
    {
      m_sendBuffer += line + "\n";
      checkEventMasks ();
    }
}

void LASCIIProtocolConnection::checkEventMasks ( void )
{
  LAutoLock _lock ( m_sendLock );

  if ( m_sendBuffer.length() )
    {
      m_fileDesc.setEventMask ( LFileDesc::eventRead | LFileDesc::eventWrite );
    }
  else
    {
      if ( m_closeMarker )
	close ();
      else
	m_fileDesc.setEventMask ( LFileDesc::eventRead );
    }
}

void LASCIIProtocolConnection::eventWrite ( LFileDesc * fd )
{
  LAutoLock _lock ( m_sendLock );

  int rc = fd->write ( (char*)m_sendBuffer, m_sendBuffer.length() );
  if ( rc>0 )
    {
      m_sendBuffer = m_sendBuffer.subString ( rc );
      checkEventMasks ();
    }
  else
    {
      close ();
    }
  // FIXME: handle error
}

void LASCIIProtocolConnection::eventRead ( LFileDesc * fd )
{
  LTRACEX_FUNC("void LASCIIProtocolConnection::eventRead ( LFileDesc * fd )" );

  LAutoLock _lock ( m_readLock );

  char buffer[4096];
  memset ( buffer, 0, sizeof(buffer) );
  int rc = fd->read ( (void*)buffer, sizeof(buffer) );

  LTRACE("LASCIIProtocolConnection.read", "read : rc = " + LString(rc) );

  if ( rc >0 )
    {
      LTRACE("LASCIIProtocolConnection.read", "Got(" + LString(rc) + "): " + buffer );
      unsigned int i = 0;
      while ( buffer[i] )
	{
	  switch ( buffer[i] )
	    {
	    case '\n':
	      if ( !m_skipLine )	
		{
		  if ( m_handler )
		    if ( (m_handler->maxMessageSize()>=m_readBuffer.length()))
		      m_handler->gotLine ( m_readBuffer );
		}
	      else
		{
		  m_skipLine = false;
		}
	      m_readBuffer = "";
	      break;
	    case '\r':
	      break;
	    default:
	      if ( !m_skipLine )
		m_readBuffer += buffer[i];
	      break;
	    }
	  i++;
	}
      if ( m_handler )
	if ( m_readBuffer.length() > m_handler->maxMessageSize() )
	  {
	    m_readBuffer = "";
	    m_skipLine = true;
	  }
    }
  else
    {
      close ();
    }
  // FIXME: handle error
}

void LASCIIProtocolConnection::eventSelectTimeout ()
{
  checkEventMasks ();
}

bool LASCIIProtocolConnection::isConnected () const
{
  return m_connected;
}

void LASCIIProtocolConnection::requestClose ( void )
{
  m_closeMarker = true;
}

void LASCIIProtocolConnection::close ( void )
{

  LAutoLock _lock_r ( m_readLock );
  LAutoLock _lock_w ( m_sendLock );

  m_readSlot.disconnect ();
  m_writeSlot.disconnect ();
  m_timeoutSlot.disconnect ();

  m_fileDesc.unreg ( LSelectLoop::getInstance() );
  m_fileDesc.close ();  
  m_connected = false;
}

void LASCIIProtocolConnection::setTimeout ( luint32 timeout )
{
  m_timeout = timeout;
}

void LASCIIProtocolConnection::tickLastData ()
{
  m_lastData = time(0);
}

void LASCIIProtocolConnection::eventTimeout ()
{
  close ();
}

bool LASCIIProtocolConnection::isTimedOut () const
{
  return time(0)-m_lastData >= m_timeout;
}

// ---------->>>> LASCIIProtocolTCPServer

LASCIIProtocolTCPServer::LASCIIProtocolTCPServer ( LManagedPointer<LASCIIProtocolFactory> factory ) :
  m_factory ( factory )
{
  m_tickerSlot = connect_emitter_0 ( &LASCIIProtocolTCPServer::removeDead, *this, LSelectLoop::getInstance()->timeoutEvent );
}

LASCIIProtocolTCPServer::~LASCIIProtocolTCPServer ()
{
  m_tickerSlot.disconnect ();
}

void LASCIIProtocolTCPServer::newConnection ( LSocket4Inet & connection )
{
  LAutoLock _lock ( m_lock );

  LManagedPointer<LASCIIProtocolConnection> aconn;
  LASCIIProtocolHandler * handler = m_factory->createHandler();
  aconn = new LASCIIProtocolConnection ( handler, connection );

  m_connList.add ( aconn );
}

void LASCIIProtocolTCPServer::removeDead ( void )
{
  LAutoLock _lock ( m_lock );

  luint32 num = m_connList.entries();

  LManagedPointer<LCursor> cursor = m_connList.createCursor ();
  LHeap < LManagedPointer < LASCIIProtocolConnection > > connList;
  forcursor ( cursor )
    {
      LManagedPointer<LASCIIProtocolConnection> conn;
      conn = m_connList.elementAt ( cursor );
      if ( conn -> isConnected () )
	{
	  if ( conn->isTimedOut () )
	    {
	      conn->eventTimeout ();
	    }
	  else
	    {
	      connList.add ( m_connList.elementAt ( cursor ) );
	    }
	}
    }
  m_connList.clear ();
  m_connList = connList;

  if ( m_connList.entries() != num )
    {
      LTRACE ( "LASCIIProtocolTCPServer", "Removed " + LString(num-m_connList.entries() ) + " connection(s)" );
    }
}
