#include <lcl/lcl_config.h>

#ifdef HAVE_LIBMYSQLCLIENT

#include <lcl/ldbsystem.mysql.h>

/*!
 * \class LDBDriver_MySQL
 * \ingroup Database
 * \brief The MySQL database driver
 * \author Jens Reimann <ctron@dentrassi.de>
 *
 * This class is the MySQL database driver. It uses (if compiled in) the mysqlclient library.
 *
 * This database driver accepts the following options:
 * user, password, database, host
 */

LDBDriver_MySQL::LDBDriver_MySQL ( LHash<LString,LString> const & vars ) :
  LDBDriver ( vars )
{
  valid = false;

  //  mysql_init ( &mysql );

  connect ();
}

LDBDriver_MySQL::~LDBDriver_MySQL ()
{
  disconnect ();
}

/*!
 * creates a new database query for the MySQL database driver
 */

LManagedPointer<LDBQuery> LDBDriver_MySQL::createQuery ()
{
  return new LDBQuery_MySQL ( *this );
}

/*!
 * \return true if the database could be opened
 *
 * Opens a new connection to a MySQL database.
 *
 * This method is normally called by ::openDatabase
 */

bool LDBDriver_MySQL::connect ()
{

  if ( connected () )
    return false;

  valid = false;

  cout << getOption("HOST") << getUser() <<  getPassword() << getDatabase() << endl;

  if ( !mysql_connect ( &mysql, getOption("HOST") , getUser(), getPassword() ) ) {
    return false;
  }

  valid = !mysql_select_db ( &mysql, getDatabase() );
  if (!valid) {
    cout << "MySQL: " << mysql_error ( &mysql ) << "\n";
    return false;
  }
  return connected();
}

/*!
 * \return true if database could be successfully closed
 *
 * This method is normally called by the destructor.
 */

bool LDBDriver_MySQL::disconnect ()
{
  if ( !connected() )
    return false;

  mysql_close ( &mysql );
  valid = false;

  return true;
}

/*!
 * \return true if the database is currently connected
 */

bool LDBDriver_MySQL::connected () const
{
  return valid;
}




LDBQuery_MySQL::LDBQuery_MySQL ( LDBDriver_MySQL & drv ) :
  LDBQuery ( drv ),
  driver ( drv )
{
  result = NULL;
  valid = false;
}

LDBQuery_MySQL::LDBQuery_MySQL ( LDBDriver_MySQL & drv, LString const & sql ) :
  LDBQuery ( drv, sql ), 
  driver ( drv )
{
  result = NULL;
  valid = false;
}

LDBQuery_MySQL::~LDBQuery_MySQL ()
{
  if (result)
    mysql_free_result ( result );
  result = NULL;
  valid = false;
}

/*!
 * \return the number of fields in the query result
 *
 * This method returns the number of fields in the query result. If the query is not valid zero will be
 * returned.
 */

long LDBQuery_MySQL::numFields () const
{
  if ( !valid || !result )
    return 0;
  
  return mysql_num_fields ( result );
}

bool LDBQuery_MySQL::exec ( LString const & sql )
{

  LTRACE ( "SQL", "Executing: '" + sql + "'");

  driver.lock.lock ();
  valid = (mysql_query ( &driver.mysql, (const char*)sql ) == 0);
  result = mysql_store_result ( &driver.mysql );
  driver.lock.unlock ();

  if ( !result )
    {
      LTRACE ( "SQLErrors", mysql_error ( &driver.mysql ) );
    }
  //cout << "MySQL: " << mysql_error ( &driver.mysql ) << "\n";

  return isValid();
}

/*!
 * \return true if the next row has been sucessfully fetched
 *
 * Using this function you may iterate through the result set. It will return \b true as long as
 * there are unfetched entries in the query result.
 *
 * The row will be stored in an internal hash and may be queried using getField!
 *
 * If you want to store the data in your own hash table you may call
 * LDBQuery_MySQL::fetchNext ( LHash<LString, LString> & hash )
 *
 * Mixing calls to the internal buffer and your own hash are allowed
 */

bool LDBQuery_MySQL::fetchNext ()
{
  if ( !isValid() || !result )
    return false;

  driver.lock.lock ();
  the_row = mysql_fetch_row ( result );
  driver.lock.unlock ();

  return the_row!=NULL;
}

/*!
 * \return true if the next row has been sucessfully fetched
 * \param hash the hash table where the row should be stored
 *
 * Using this function you may iterate through the result set. It will return \b true as long as
 * there are unfetched entries in the query result.
 *
 * The row will be stored in the hash table provided. You may access the row using the methods of LHash.
 * The key will always be the name of the field. The value the actual data of the field.
 *
 */

bool LDBQuery_MySQL::fetchNext ( LHash<LString, LString> & hash )
{
  long i;

  if ( !isValid() || !result )
    return false;

  bool flag = fetchNext ();

  if ( flag )
    for ( i = 0; i<numFields(); i++ ) {
      hash.set ( getFieldName ( i ), getField ( i ) );
    }

  return flag;
}

/*!
 * \return the data of the field
 * \param the number of the field
 *
 * You wil receive the value of this field using this method. The field number may be looked up using getField
 *
 * This works only if you read in the last row using LDBQuery_MySQL::fetchNext ()
 */

LString LDBQuery_MySQL::getField ( long i ) const
{
  if ( !isValid() || !result  )
    return LString();

  if ( i < 0 )
    return LString();

  if ( i < numFields() ) {
    return the_row[i];
  }
  return LString();
}

/*!
 * \return the data of the field
 * \param the name of the field
 *
 * You wil receive the value of this field using this method. The field name may be looked up using getField
 *
 * This works only if you read in the last row using LDBQuery_MySQL::fetchNext ()
 */

LString LDBQuery_MySQL::getField ( LString const& name ) const
{
  return getField ( getFieldIdx ( name ) );
}

/*!
 * \return the name of the field
 * \param the number of the field
 *
 * You can look up the name of a field by number using this method
 */

LString LDBQuery_MySQL::getFieldName ( long i ) const
{
  if ( !isValid() || !result )
    return LString();
  if ( i > numFields() || i < 0 )
    return LString();

  MYSQL_FIELD * field;

  driver.lock.lock ();
  mysql_field_seek ( result, i ); 
  field = mysql_fetch_field ( result );
  driver.lock.unlock ();
  return field->name;
}

/*!
 * \return the number of the field
 * \param the name of the field
 *
 * You can look up the number of a field by name using this method
 */

long LDBQuery_MySQL::getFieldIdx ( LString const & name ) const
{
  if ( !isValid() || !result )
    return -1;

  MYSQL_FIELD * field;

  driver.lock.lock ();
  for ( int i = 0; i<numFields(); i++ ) {
    mysql_field_seek ( result, i);
    field = mysql_fetch_field ( result );
    if ( name == field->name ) {
      driver.lock.unlock ();
      return i;
    }
  }

  driver.lock.unlock ();
  return -1;
}

#endif /*  HAVE_LIBMYSQL */
