#include <lcl/lcl_config.h>

#include <lcl/lfdlayer.win32.h>
#include <lcl/lfdlayer.common.h>
#include <lcl/ltrace.h>
#include <lcl/lstreamproxy.h>
#include <lcl/lrevision.h>
#include <lcl/lwin32.fd_socket.h>
#include <lcl/lwin32.fd_file.h>
#include <lcl/lpointer.h>
#include <lcl/lwin32.select.h>
#include <lcl/llogsystem.h>
#include <lcl/lmutex.h>
#include <lcl/lapp.h>

#include <iostream.h>

LTRACE_CHANNEL_STATIC("LFDManager", __lcl_trace_fdmanager_1, "File Descriptor Manager" );

#define FD2HANDLE(f, h) \
  fdHashLock.lock (); \
  if ( !fdHash.isSet ( f) ) { \
    LTRACE ( "LFDManager", "Invalid Handle: " + LString(f) ); \
    fdHashLock.unlock (); \
    throw LFDManagerException ( __FILE__, __LINE__, LFDManagerException::invalidFD, f ); \
    } \
  h = fdHash.get ( f ); \
  fdHashLock.unlock ();

#define X__FD2WIN32SELECT(f,h) \
  if ( !fdHash.isSet ( f ) ) \
    throw LFDManagerException ( __FILE__, __LINE__, LFDManagerException::invalidFD ); \
  h = fdHash.get ( f )->getSelectHandle ();

LFEATURE("LFDManager", fdm_f1, "Win32 Layer" );
LVCS_INIT_D(LFDManager, "WIN32");

LFDManager * LFDManager::theInstance = NULL;
LMutex LFDManager::instanceLock;

LFDManager * LFDManager::getInstance ()
{
  // Cannot use LAutoLock here since this would cause an inifite recursion
  instanceLock.lock ();

  if ( !theInstance )
    theInstance = new LFDManager ();

  instanceLock.unlock ();

  return theInstance;
}

LFileDescriptor LFDManager::registerHandle ( LManagedPointer<LFDManager_Basic> handle )
{
  LAutoLock _lock ( fdHashLock );

  LFileDescriptor fd = (LFileDescriptor)(rand()*rand());
  while ( fdHash.isSet ( fd ) ) {
    fd = (LFileDescriptor)(rand()*rand());
  }
  LTRACE ( "LFDManager", "Registered handle " + LString(fd) );
  fdHash.set ( fd, handle );
  return fd;
}

void LFDManager::unregisterHandle ( LFileDescriptor fd )
{
  LAutoLock _lock ( fdHashLock );
  fdHash.unset ( fd );
}

LManagedPointer<LFDManager_Basic> LFDManager::getHandler ( LFileDescriptor fd ) const
{
  LAutoLock _lock ( fdHashLock );

  if ( !fdHash.isSet ( fd ) )
    throw LFDManagerException ( __FILE__, __LINE__, LFDManagerException::invalidFD, fd );
  return fdHash.get ( fd );
}

LFDManager::LFDManager ()
{
  if ( theApp.isParam ( "lcl-win32-console" ) )
    AllocConsole ();

  lout << "LFDManager -> Creating new LCL File Descriptor Manager for WIN32!" << endl << flush;

  LFDManager_Basic * handle;
  LString temp;

  lout << "LFDManager -> Injecting STD handles ..." << endl << flush;

  handle = new LFDManager_Basic ( GetStdHandle ( STD_INPUT_HANDLE ) );
  fdHash.set ( 0, handle );

  handle = new LFDManager_Basic ( GetStdHandle ( STD_OUTPUT_HANDLE ) );
  fdHash.set ( 1, handle );
  temp = "   -> Injected STDOUT\n\r";
  handle->write ( (char const*)temp, temp.length() );

  handle = new LFDManager_Basic ( GetStdHandle ( STD_ERROR_HANDLE ) );
  fdHash.set ( 2, handle );
  temp = "   -> Injected STDERR\n\r";
  handle->write ( (char const*)temp, temp.length() );

  // Register winsock

  WORD wVersionRequested;
  WSADATA wsaData;
  int err;
 
  wVersionRequested = MAKEWORD( 2, 0 );
 
  err = WSAStartup( wVersionRequested, &wsaData );
  if ( err != 0 ) {
    LLOG ( LLogSystem::FATAL, "WinSock is needed!!" ); 
    exit(1);
  }
 
  /* Confirm that the WinSock DLL supports 2.0.*/
  /* Note that if the DLL supports versions greater    */
  /* than 2.0 in addition to 2.0, it will still return */
  /* 2.0 in wVersion since that is the version we      */
  /* requested.                                        */
 
  if ( LOBYTE( wsaData.wVersion ) != 2 ||
       HIBYTE( wsaData.wVersion ) != 0 ) {
    LLOG ( LLogSystem::FATAL, "WinSock2 is needed!!" ); 
    WSACleanup( );
    exit(1);
  }
 
}

bool LFDManager::is_valid ( LFileDescriptor fd ) const
{
  LTRACE_FUNC("LFDManager", "is_valid");

  if ( fd == -1 )
    {
      LLOG ( LLogSystem::INFO, "Called isValid() with fd=-1" );
      return false;
    }

  LAutoLock _lock ( fdHashLock );
  if ( !fdHash.isSet ( fd ) )
    return false;

  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->isValid ();
}

bool LFDManager::is_socket ( LFileDescriptor fd ) const
{

  if ( fd == -1 )
    {
      LLOG ( LLogSystem::WARNING, "Called isSocket() with fd=-1" );
      return false;
    }

  LAutoLock _lock ( fdHashLock );

  if ( !fdHash.isSet ( fd ) )
    return false;
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->isSocket ();
}

int LFDManager::sys_listen ( LFileDescriptor fd, int backlog )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->listen ( backlog );
}

int LFDManager::sys_connect ( LFileDescriptor fd, struct sockaddr const * addr, unsigned long addr_len )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->connect ( addr, addr_len );
}

int LFDManager::sys_bind ( LFileDescriptor fd, struct sockaddr const * addr, unsigned long addr_len )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->bind ( addr, addr_len );
}

LFileDescriptor LFDManager::sys_accept ( LFileDescriptor fd, struct sockaddr * addr, unsigned long * addr_len )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  LFDManager_Basic * new_handle = handle->accept ( addr, addr_len );
  
  if ( new_handle == NULL )
    return -1;
  else
    return registerHandle ( new_handle );
}

int LFDManager::sys_sendto ( LFileDescriptor fd, void const * data, unsigned long len, int flags, struct sockaddr const *to, unsigned long tolen )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->sendto ( data, len, flags, to, tolen );
}

int LFDManager::sys_recvfrom ( LFileDescriptor fd, void * data, unsigned long len, int flags, struct sockaddr *from, unsigned long * fromlen )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->recvfrom ( data, len, flags, from, fromlen );
}

int LFDManager::sys_setsockopt ( LFileDescriptor fd, int level, int optname, void const * optval, unsigned long optlen )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->setsockopt ( level, optname, optval, optlen );
}

int LFDManager::sys_getsockopt ( LFileDescriptor fd, int level, int optname, void * optval, unsigned long *optlen )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->getsockopt ( level, optname, optval, optlen );
}


int LFDManager::sys_getsockname ( LFileDescriptor fd, struct sockaddr *name, unsigned long *namelen )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->getsockname ( name, namelen );
}

int LFDManager::sys_getpeername ( LFileDescriptor fd, struct sockaddr *name, unsigned long *namelen )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->getpeername ( name, namelen );
}

int LFDManager::sys_read ( LFileDescriptor fd, void * ptr, unsigned long size )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->read ( ptr, size );
}

int LFDManager::sys_write ( LFileDescriptor fd, void const * ptr, unsigned long size )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->write ( ptr, size );
}

int LFDManager::sys_close ( LFileDescriptor fd )
{
  if ( fd == -1 )
    {
      LLOG ( LLogSystem::WARNING, "Called close() with fd=-1" );
      return -1;
    }

  if ( !is_valid ( fd ) )
    return -1;

  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  int rc = handle->close ( );

  unregisterHandle ( fd );

  return rc;
}

LFileDescriptor LFDManager::sys_dup ( LFileDescriptor fd )
{
  if ( !is_valid ( fd ) )
    {
      LTRACE ( "LFDManager", "called dup() with an invalid fd=" + LString(fd) );
      return -1;
    }

  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  LFDManager_Basic * new_fd = handle->dup ();
  if ( new_fd == NULL )
    {
      LTRACE ( "LFDManager", "duplicate failed (fd=" + LString(fd) + ") :" + LString::fromErrorCode() );
      return -1;
    }
  return registerHandle ( new_fd );
}

LFileDescriptor LFDManager::sys_open ( char const * filename, int accessType, luint32 om )
{
  LFDManager_File * file = new LFDManager_File;

  if ( !file->open ( filename, accessType, om ) )
    return -1;

  return registerHandle ( file );
}

LFileDescriptor LFDManager::sys_socket ( int af, int type, int proto )
{
  LFDManager_Socket * socket = new LFDManager_Socket;
  if ( !socket->socket ( af, type, proto ) )
    return -1;

  return registerHandle ( socket );
}

bool LFDManager::sys_getblocking ( LFileDescriptor fd )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  return handle->getBlocking ();
}

void LFDManager::sys_setblocking ( LFileDescriptor fd, bool state )
{
  LFDManager_Basic * handle;
  FD2HANDLE(fd,handle);

  handle->setBlocking ( state );
}


bool LFDManager::sys_select (
			     LSet<LFileDescriptor> const & read,
			     LSet<LFileDescriptor> const & write,
			     LSet<LFileDescriptor> const & xcpt,
			     LSet<LFileDescriptor> & read_res,
			     LSet<LFileDescriptor> & write_res,
			     LSet<LFileDescriptor> & xcpt_res,
			     unsigned long msecs
			     )
{
  //  LWin32Select select;
  return LWin32Select::sys_select ( read, write, xcpt, read_res, write_res, xcpt_res, msecs );
}


