#include <lcl/lfdsendqueue.h>
#include <lcl/ltrace.h>
#include <lcl/lstreamproxy.h>

LVCS_INIT(LFDSendQueue);


LFDSendQueueStream::LFDSendQueueStream ( LManagedPointer<LFDSendQueueStorage> _storage ) :
  storage ( _storage )
{
  LTRACE_FUNC ( "LFDSendQueueStream", "LFDSendQueueStream" );
}

LFDSendQueueStream::~LFDSendQueueStream ()
{
  LTRACE_FUNC ( "LFDSendQueueStream", "~LFDSendQueueStream" );
}

bool LFDSendQueueStream::mayWrite ()
{
  return storage->isValid ();
}

int LFDSendQueueStream::write ( char c )
{
  LTRACE_FUNC ( "LFDSendQueueStream", "write(char)" );

  if ( !mayWrite() ) {
    LTRACE ( "LFDSendQueueStream", "Not allowed to write!" );
    return 0;
  }

  storage->append ( &c, 1 );

  return 1;
}

int LFDSendQueueStream::write ( void const * ptr, unsigned long len )
{
  LTRACE_FUNC ( "LFDSendQueueStream", "write(void const*, unsigned long)" );
  if ( !mayWrite() ) {
    LTRACE ( "LFDSendQueueStream", "Not allowed to write!" );
    return 0;
  }

  storage->append ( ptr, len );

  return len;
}

void LFDSendQueueStream::close ( )
{
  LTRACE_FUNC ( "LFDSendQueueStream", "close" );
  storage->close ();
}

LManagedPointer<LFDSendQueueStorage> LFDSendQueue::getStorage ( LReference<LFileDesc> fd )
{
  if ( !fdHash.isSet ( fd ) ) {
    LTRACE ( "LFDSendQueue", "Creating new storage for " + LString((int)fd->getHandle()) );
    fdHash.set ( fd, new LFDSendQueueStorage ( *this, fd ) );        

    connect_emitter_1 ( &LFDSendQueue::afterWriteEvent, *this, fd->writeEvent );
  }

  return fdHash.get ( fd );
}

LSelectLoop * LFDSendQueue::getLoop () const
{
  return LSelectLoop::getInstance ();
}

void LFDSendQueue::afterWriteEvent ( LFileDesc * fd )
{
  LTRACE_FUNC ( "LFDSendQueue", "afterWriteEvent" );
  cleanUp ();
}

void LFDSendQueue::cleanUp ()
{
  LFDSendQueue::CleanUpIter cui;
  fdHash.allElementsDo ( cui );

  LCursor * cursor = cui.cleanList.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() ) {
    LTRACE ( "LFDSendQueue", "remove element for fd " + LString((int)cui.cleanList.elementAt (cursor)->getHandle() ) );
    if ( !fdHash.isSet ( cui.cleanList.elementAt ( cursor ) ) )
      LTRACE ( "LFDSendQueue", "Points to non-hash element" );
    fdHash.unset ( cui.cleanList.elementAt ( cursor ) );
  }
  delete cursor;
}

void LFDSendQueue::CleanUpIter::applyTo ( LReference < LFileDesc > & fd, LManagedPointer<LFDSendQueueStorage> & fs)
{
  if ( !fs->isValid() )
    cleanList.add ( fd );
}

LFDSendQueueStorage::LFDSendQueueStorage ( LFDSendQueue & queue, LReference<LFileDesc> _fd ) :
  myQueue ( queue ),
  fd ( _fd ),
  valid ( true ),
  closed ( false )
{
  fd->setEventMask ( fd->eventMask() | LFileDesc::eventWrite );
  connect_emitter_1 ( &LFDSendQueueStorage::processSend, *this, fd->writeEvent );

  lout << "Create send queue for " << (long int)fd->getHandle() << endl << flush;
}

LFDSendQueueStorage::~LFDSendQueueStorage ()
{
  lout << "Destroy send queue for " << (long int)fd->getHandle() << endl << flush;
}

void LFDSendQueueStorage::append ( void const * ptr, unsigned long len )
{
  if ( closed )
    return;

  LTRACE ( "LFDSendQueueStorage", "appending " + LString(len) + " bytes!" );

  buffer.append ( ptr, len );
  if ( len )
    {
      LTRACE ( "LFDSendQueueStorage", "buffer contains data, registering for select loop" );
      fd->reg ( myQueue.getLoop() );
    }
}

void LFDSendQueueStorage::processSend ( LFileDesc * _fd )
{
  LTRACE_FUNC ( "LFDSendQueueStorage", "processSend" );

  int rc = fd->write ( buffer.getConstPtr(), buffer.getSize() );
  if ( rc > 0 ) {
    buffer.cutFrom ( rc );
  } else {
    closed = valid = false;
    buffer = LBuffer();
  }
  if ( !buffer.getSize () ) {
    fd->unreg ( myQueue.getLoop () );
    if ( closed )
      valid = false;
  }
}

void LFDSendQueueStorage::close ( )
{
  closed = true;
}

bool LFDSendQueueStorage::isValid ( ) const
{
  return valid;
}

bool LFDSendQueueStorage::isClosed ( ) const
{
  return closed;
}
