#include <lcl/lfilewatcher.h>

#include <lcl/lpointer.h>
#include <lcl/ltrace.h>

LFileWatcher::LFileWatcher () :
  runnerFn ( &LFileWatcher::runner, *this )
{
}

void LFileWatcher::runner ( void )
{
  while ( true )
    {
      checkOnce ();
      sleep ( 1 );
    }
}

/*!
 * Check all files once.
 *
 * This function will return immediately after checking all files.
 *
 * \note All events will be fired before the function returns.
 * 
 */

void LFileWatcher::checkOnce ( void )
{
  LAutoLock autoLock ( lock );

  LManagedPointer<LCursor> cursor = fileList.createCursor ();

  forcursor ( cursor )
    {
      checkFile ( fileList.elementAt ( cursor ) );
    }
}

void LFileWatcher::checkFile ( FileInfo & fi )
{
  LTRACE ( "LFileWatcher", "Checking " + fi.fileName );
  struct stat si;
  int rc = stat ( fi.fileName, &si );

  if ( !fi.checked )
    {
      fi.checked = true;
    }
  else
    {
      if ( fi.last_valid != (rc==0) )
	{
	  changed.emit ( fi.fileName, valid );
	  changedStatus.emit ( fi.fileName );	  
	}
      if ( rc == 0 )
	{
	  if ( fi.last_size != si.st_size )
	    {
	      changed.emit ( fi.fileName, size );
	      changedSize.emit ( fi.fileName );
	    }
	  if ( fi.last_mtime != si.st_mtime )
	    {
	      changed.emit ( fi.fileName, mtime );
	      changedMTime.emit ( fi.fileName );
	    }
	  if ( fi.last_atime != si.st_atime )
	    {
	      changed.emit ( fi.fileName, atime );
	    }
	  if ( fi.last_ctime != si.st_ctime )
	    {
	      changed.emit ( fi.fileName, ctime );
	    }
	}
    }
  fi.last_valid = ( rc == 0 );
  if ( fi.last_valid )
    {
      fi.last_size = si.st_size;
      fi.last_mtime = si.st_mtime;
      fi.last_atime = si.st_atime;
      fi.last_ctime = si.st_ctime;
    }

  LTRACE ( "LFileWatcher", "Size: " + LString(fi.last_size) + " -> " + LString(si.st_size) );
}

/*!
 *
 * Adds a new file to the list of files beeing check by LFileWatcher.
 *
 * \note At the first run there won't be any events fired since this. So the initial
 * status is the status which the file has at the first check.
 *
 * \param filename the path of the file to be added
 */

void LFileWatcher::addFile ( LString const & filename )
{

  LAutoLock autoLock ( lock );

  FileInfo fi;
  fi.fileName = filename;
  fi.checked = false;

  fileList.add ( fi );
}

/*!
 * Remove a file from the list of checked files.
 */
void LFileWatcher::removeFile ( LString const & filename )
{

  LAutoLock autoLock ( lock );

  fileList.removeElementWithKey ( filename );
}

void LFileWatcher::start ( void )
{
  LThread tco ( &runnerFn );
  tco.start ();
}

LCL_EXPORT LString key ( LFileWatcher::FileInfo const & p1 )
{
  return p1.fileName;
}

