#include <lcl/limage.bmp.h>

#include <lcl/lstringtokenizer.h>
#include <lcl/ltrace.h>

#include <lcl/lstreamproxy.h>

typedef struct {
  unsigned short magic_id;
  unsigned long size;

  unsigned short dummy1, dummy2;

  unsigned long offsetBits;
} BitmapHeader;

typedef struct {
  unsigned long size;
  long width, height;
  unsigned short planes, bpp;
  unsigned long compression;
  unsigned long img_size;
  long pels_x, pels_y;
  unsigned long colors, color_imp;
  
} BitmapInfo;

typedef struct {
  unsigned char red, green, blue, dummy;
} BitmapRGBX;

bool LImageReader_BMP::impl_read ( LImage  & image, LIStream & stream ) const
{
  BitmapHeader header;
  BitmapInfo info;
  LColorMap cMap;

  stream.read ( &header, sizeof(header) );
  if ( header.magic_id !=  ( (((int)'B') << 8) | ((int)'M') ) ) {
    LTRACE ( "LImageReader_BMP", "magic id is not 'BM'" );
    return false;
  }

  stream.read ( &header, sizeof(info) );
  image.setSize ( info.width, info.height );
  image.setBitsPerPixel ( info.bpp );

  unsigned long num_colors = 0;
  BitmapRGBX * colorMap = NULL;

  if ( info.colors ) {
    num_colors = info.colors;
  } else if ( info.bpp <= 8 ) {
    num_colors = 1 << (info.bpp-1);
  }

  if ( num_colors ) {
    colorMap = new BitmapRGBX [num_colors];
    stream.read ( colorMap, sizeof(colorMap)*num_colors );   
    cMap = LColorMap ( num_colors );
    int i;
    for ( i = 0; i<num_colors; i++ ) {
      cMap[i] = LColor ( (char)colorMap[i].red, (char)colorMap[i].green, (char)colorMap[i].blue, ( char)0 );      
    }
    delete [] colorMap;
    colorMap = NULL;
    image.setColorMap ( cMap );
  }

  switch ( info.bpp ) {
  case 8:
    return readBits_8 ( image, stream, info.width, info.height );
  case 24:
    return readBits_24 ( image, stream, info.width, info.height );
  case 32:
    return readBits_32 ( image, stream, info.width, info.height );
  default:
    return false;
  }

  //  return true;
}

bool LImageReader_BMP::readBits_8 ( LImage & image, LIStream & stream, long width, long height ) const
{
  char * bitSpace = (char*)malloc ( 1 * width * height );

  unsigned long line_size = 1 * width;
  unsigned long line_delta = 0;

  if ( line_size % 4 ) {
    line_size += 4 - (line_size % 4);
    line_delta = line_size - ( 1*width );
  }

  unsigned long file_size = line_size * height;

  char * fileSpace = (char*)malloc ( file_size );

  stream.read ( fileSpace, file_size );
  
  // BPP Specific code ...
  long pos1, pos2;
  for ( int iy = 0; iy < height; iy++ ) {
    for ( int ix = 0; ix < width; ix++ ) {
      pos1 = (ix + (iy*width)) * 1;
      pos2 = ((ix + (((height-1)-iy)*width)) * 1) + (((height-1)-iy) * line_delta);
      bitSpace [pos1] = fileSpace[pos2];
    }
  }
  // BPP Specific code end!!

  image.setBits ( LBuffer ( 1 * width * height, bitSpace ) );

  free ( fileSpace );
  free ( bitSpace );

  return true;
}


bool LImageReader_BMP::readBits_24 ( LImage & image, LIStream & stream, long width, long height ) const
{
  char * bitSpace = (char*)malloc ( 3 * width * height );

  unsigned long line_size = 3 * width;
  unsigned long line_delta = 0;

  if ( line_size % 4 ) {
    line_size += 4 - (line_size % 4);
    line_delta = line_size - ( 3*width );
  }

  unsigned long file_size = line_size * height;

  char * fileSpace = (char*)malloc ( file_size );

  stream.read ( fileSpace, file_size );
  
  // BPP Specific code ...
  long pos1, pos2;
  for ( int iy = 0; iy < height; iy++ ) {
    for ( int ix = 0; ix < width; ix++ ) {
      pos1 = (ix + (iy*width)) * 3;
      pos2 = ((ix + (((height-1)-iy)*width)) * 3) + (((height-1)-iy) * line_delta);
      bitSpace [pos1] = fileSpace[pos2];
      bitSpace [pos1+1] = fileSpace[pos2+1];
      bitSpace [pos1+2] = fileSpace[pos2+2];
    }
  }
  // BPP Specific code end!!

  image.setBits ( LBuffer ( 3 * width * height, bitSpace ) );

  free ( fileSpace );
  free ( bitSpace );

  return true;
}

bool LImageReader_BMP::readBits_32 ( LImage & image, LIStream & stream, long width, long height ) const
{
  char * bitSpace = (char*)malloc ( 4 * width * height );

  unsigned long line_size = 4 * width;
  unsigned long line_delta = 0;

  if ( line_size % 4 ) {
    line_size += 4 - (line_size % 4);
    line_delta = line_size - ( 4*width );
  }

  unsigned long file_size = line_size * height;

  char * fileSpace = (char*)malloc ( file_size );

  stream.read ( fileSpace, file_size );
  
  // BPP Specific code ...
  long pos1, pos2;
  for ( int iy = 0; iy < height; iy++ ) {
    for ( int ix = 0; ix < width; ix++ ) {
      pos1 = (ix + (iy*width)) * 4;
      pos2 = ((ix + (((height-1)-iy)*width)) * 4) + (((height-1)-iy) * line_delta);
      bitSpace [pos1] = fileSpace[pos2];
      bitSpace [pos1+1] = fileSpace[pos2+1];
      bitSpace [pos1+2] = fileSpace[pos2+2];
      bitSpace [pos1+3] = 0;
    }
  }
  // BPP Specific code end!!

  image.setBits ( LBuffer ( 4 * width * height, bitSpace ) );

  free ( fileSpace );
  free ( bitSpace );

  return true;
}
