#include <lcl/limage.h>

#include <lcl.h>

LColorMap::LColorMap () :
  LTable<LColor> ()
{
}

LColorMap::LColorMap ( LColorMap const & p1 ) :
  LTable<LColor> ( p1 )
{
}

LColorMap::LColorMap ( unsigned long num_colors ) :
  LTable<LColor> ( num_colors )
{
}

unsigned long LColorMap::findColor ( LColor const & p1 )
{
  unsigned long i;
  for ( i = 0; i<getSize(); i++ ) {
    if ( (*this)[i] == p1 )
      return i;
  }
  return 0xFFFFFFFF;
}

bool LColorMap::containsColor ( LColor const & p1 )
{
  return findColor(p1) != 0xFFFFFFFF;
}

LImage::LImage () :
  width ( 0 ),
  height ( 0 ),
  bpp ( 0 ),
  valid ( false )
{
}

LImage::LImage ( unsigned long _width, unsigned long _height, unsigned long _bpp ) :
  width ( _width ),
  height ( _height ),
  bpp ( _bpp ),
  bits ( width*height*(bpp/8) ),
  valid ( true )
{
}

LImage::LImage ( unsigned long _width, unsigned long _height, unsigned long _bpp, LColorMap const & _map ) :
  width ( _width ),
  height ( _height ),
  bpp ( _bpp ),
  bits ( width*height*(bpp/8) ),
  colorMap ( _map )
{
}

void LImage::setColorMap ( LColorMap const & _map )
{
  colorMap = _map;
}

void LImage::setBitsPerPixel ( unsigned long _bpp )
{
  bpp = _bpp;
  bits.resize ( width*height*(bpp/8) );
  bits.erase ();
}

void LImage::setSize ( unsigned long _width, unsigned long _height )
{
  width = _width;
  height = _height;
  bits.resize ( width*height*(bpp/8) );
  bits.erase ();
}

unsigned long LImage::getHeight () const
{
  return height;
}

unsigned long LImage::getWidth () const
{
  return width;
}

unsigned long LImage::getBitsPerPixel () const
{
  return bpp;
}

void LImage::setBits ( LBuffer const & _bits )
{
  unsigned long len = MIN ( bits.getSize(), _bits.getSize() );
  memcpy ( bits.getPtr(), _bits.getConstPtr(), len );
}

LColor LImage::getPixel ( unsigned long x, unsigned long y ) const
{
  if ( (x >= width) || (y >= height) )
    return LColor ();

  unsigned long pos = (x + (y*width)) * (bpp/8);
  unsigned long t1, t2, t3, t4;

  switch ( bpp ) {
  case 1:
    t1 = x % 8;
    t2 = 1 << t1;
    t3 = bits[pos] & t2;
    return t3 ? LColor() : LColor(0xFFFFFFFFL,0xFFFFFFFFL,0xFFFFFFFFL,0UL );

  case 4:
    t1 = (x % 2) * 4;
    t2 = bits[pos] >> t1;
    if ( colorMap.getSize() )
      return colorMap[t2];
    else
      return LColor ( (int)t2<<4, (int)t2<<4, (int)t2<<4, (int)0 );

  case 8:
    t1 = bits[pos];
    if ( colorMap.getSize() )
      return colorMap[t1];
    else
      return LColor ( (unsigned char)t1, (unsigned char)t1, (unsigned char)t1, (unsigned char)0 );

  case 16:
    t1 = bits[pos];
    t2 = bits[pos+1];
    return LColor (
		   (unsigned char)((t1 & 0xF8) >> 3),
		   (unsigned char)( ( (t1 & 0x07) << 2) | ( (t2 & 0xC0) >> 6 ) ),
		   (unsigned char)( ( t2 & 0x3F >> 1 ) ),
		   (unsigned char)0
		   );

  case 24:
    return LColor ( (unsigned char)bits[pos], (unsigned char)bits[pos+1], (unsigned char)bits[pos+2], (unsigned char)0 );
 
  case 32:
    return LColor ( (unsigned char)bits[pos], (unsigned char)bits[pos+1], (unsigned char)bits[pos+2], (unsigned char)bits[pos+3] );

  default:
    return LColor();
  }
}

void LImage::putPixel ( unsigned long x, unsigned long y, LColor const & _color )
{
  if ( (x >= width) || (y >= height) )
    return;

  LColor color = getPixel ( x, y );
  LColor new_color = _color;

  unsigned long pos = (x + (y*width)) * (bpp/8);
  unsigned long t1, t2, t3, t4;

  switch ( bpp ) {
  case 1:
    t1 = x % 8;
    t2 = 1 << t1;
    if ( new_color == LColor() )
      bits[pos] = bits[pos] & (~t2);
    else
      bits[pos] = bits[pos] | t2;
    break;

  case 4:
    t1 = (x % 2) * 4;
    t2 = bits[pos] >> t1;
    if ( colorMap.getSize() ) {
      t2 = colorMap.findColor(new_color);
      if ( t2 == 0xFFFFFFFF )
	return;
    } else {
      new_color.toGray();
      t2 = new_color.getBlueAsChar() >> 4;      
    }

    // clear the 4 bits
    if ( t1 )
      bits[pos] = bits[pos] & 0xF0;
    else
      bits[pos] = bits[pos] & 0x0F;
    
    // and set them
    bits[pos] = (bits[pos] << t1) | t2;
    break;

  case 8:
    if ( colorMap.getSize() ) {
      t1 = colorMap.findColor(new_color);
      if ( t1 == 0xFFFFFFFF )
	return;
      bits[pos] = t1;
    } else {
      new_color.toGray ();
      bits[pos] = new_color.getBlueAsChar();
    }
    break;

  case 16:
    t1 = 0;
    t2 = 0;

    t1 = (new_color.getRedAsChar() >> 3) << 3;
    t1 = t1 | (new_color.getGreenAsChar() >> 5 );
    t2 = (new_color.getGreenAsChar() >> 2) & 0x03;
    t2 = t2 | ((new_color.getBlueAsChar() >> 2) << 1 );
    
    bits[pos] = t1;
    bits[pos+1] = t2;
    break;

  case 24:
    bits[pos] = new_color.getRedAsChar();
    bits[pos+1] = new_color.getGreenAsChar();
    bits[pos+2] = new_color.getBlueAsChar();
    break;
 
  case 32:
    bits[pos] = new_color.getRedAsChar();
    bits[pos+1] = new_color.getGreenAsChar();
    bits[pos+2] = new_color.getBlueAsChar();
    bits[pos+3] = new_color.getAlphaAsChar();
    break;

  }
  
}

LColorMap & LImage::getColorMap ()
{
  return colorMap;
}

LColorMap const & LImage::getColorMap () const
{
  return colorMap;
}

bool LImage::read ( LImageReader const & reader, LIStream & ins )
{
  return reader.read ( *this, ins );
}

bool LImage::write ( LImageWriter const & writer, LOStream & outs ) const
{
  return writer.write ( *this, outs );
}

bool LImage::isValid () const
{
  return valid;
}
