#include <lcl/linterpreter.h>

#include <lcl/lstreamproxy.h>
#include <lcl/ltrace.h>

LVCS_INIT(LInterpreter);

// ========>>>>>>>>>>>> Exception

LInterpreter::Exception::Exception ( LInterpreter::Exception::Reason _reason, LString const & _tmp ) :
  LBasicException ( __FILE__, __LINE__, "Interpreter Exception", LCL_ERROR_CODE("SCR", 0x00000001) ),
  reason ( _reason ),
  tmp ( _tmp )
{
}

LInterpreter::Exception::~Exception()
{
}

LString LInterpreter::Exception::getArg1 ( ) const
{
  return tmp;
}

void LInterpreter::Exception::log () const
{
  LBasicException::log ();
  lerr << " >> LCL Interpreter Exception" << endl;
  lerr << "  >> Reason: ";

  switch ( reason )
    {
    case CMD_INVALID:
      lerr << "Unkown command: " << getArg1() << endl;
      break;
    case PARAM_CNT:
      lerr << "Wrong parameter count: " << getArg1() << endl;
      break;
    default:
      lerr << "<Unkown>" << endl;
      break;
    }
  lerr << flush;
}

// ========>>>>>>>>>>>> Sequence

void LInterpreter::Sequence::addParameter ( LManagedPointer<Sequence> _arg )
{
  args.add ( _arg );
}

LInterpreter::Sequence::~Sequence ()
{
}

// ========>>>>>>>>>>>> Sequence CONST

/*!
 * \param s the character to scan for
 * \param e the escaped character which is inserted instead
 *
 * do the const replacement code for one character in LInterpreter::Sequence_Const::Sequence_Const
 */
#define DO_ESCAPED(s,e) case s: if ( escaped ) value += e; else value+= s; escaped = false; break;

LInterpreter::Sequence_Const::~Sequence_Const ()
{
}

LInterpreter::Sequence_Const::Sequence_Const ( LVariant const & _value ) :
  value ( _value )
{
  // FIXME: re-implement backslashes ... but in parser!
  /*
  LString tmp;
  tmp = value.asString();
  unsigned long i;
  bool escaped = false;
  value = "";
  for ( i = 0; i<tmp.length(); i++ )
    {
      switch ( tmp[i] )
	{
	case '\\':
	  if ( escaped )
	    {
	      value += '\\';
	      escaped = false;
	    }
	  else
	    escaped = true;
	  break;

	  DO_ESCAPED('n', '\n');
	  DO_ESCAPED('r', '\r');

	default:
	  value += tmp[i];
	  escaped = false;
	  break;
	}
    }
  */
}

/*!
 * \return the result of the execution
 * \param inter the interpreter which runs the sequence
 *
 * Execute a Sequence_Const command. If fact this only returns the value
 * since this is a \em const command
 */
LVariant LInterpreter::Sequence_Const::exec ( LInterpreter::Context & context )
{
  return value;
}

// ========>>>>>>>>>>>> Sequence IF

LInterpreter::Sequence_If::~Sequence_If ()
{
}


LVariant LInterpreter::Sequence_If::exec ( LInterpreter::Context & context )
{
  if ( (args.entries() != 2) && (args.entries() != 3 ) )
    throw Exception ( Exception::PARAM_CNT, "if switch has 2 or 3 parameters" );

  LManagedPointer<LCursor> cursor = args.createCursor ();

  cursor->setToFirst();
  LManagedPointer<Sequence> ifCommand = args.elementAt ( cursor );

  cursor->setToNext();
  LManagedPointer<Sequence> trueCommand = args.elementAt ( cursor );

  LManagedPointer<Sequence> falseCommand;
  cursor->setToNext();
  if ( cursor->isValid() )
   falseCommand = args.elementAt ( cursor );

  if ( (bool)ifCommand->exec ( context ) )
    return trueCommand->exec ( context );
  else
    if ( falseCommand.isValid() )
      return falseCommand->exec ( context );
    else
      return "";
}

// ========>>>>>>>>>>>> Sequence FOR


LInterpreter::Sequence_For::~Sequence_For ()
{
}

LVariant LInterpreter::Sequence_For::exec ( LInterpreter::Context & context )
{
  LManagedPointer<LCursor> cursor = args.createCursor ();

  if ( args.entries() != 4 )
    throw Exception ( Exception::PARAM_CNT, "for loop needs 4 parameters" );

  cursor->setToFirst();
  LManagedPointer<Sequence> initCommand = args.elementAt ( cursor );

  cursor->setToNext();
  LManagedPointer<Sequence> validCommand = args.elementAt ( cursor );

  cursor->setToNext();
  LManagedPointer<Sequence> stepCommand = args.elementAt ( cursor );

  cursor->setToNext();
  LManagedPointer<Sequence> doCommand = args.elementAt ( cursor );

  LVariant rc;

  for ( initCommand->exec ( context ); (bool)validCommand->exec( context ); stepCommand->exec( context ) )
    {
      rc = doCommand->exec ( context );
    }

  return rc;
}


// ========>>>>>>>>>>>> Sequence WHILE

LInterpreter::Sequence_While::~Sequence_While ()
{
}

LVariant LInterpreter::Sequence_While::exec ( LInterpreter::Context & context )
{

  if ( args.entries() != 2 )
    throw Exception ( Exception::PARAM_CNT, "while loop needs 2 parameters" );


  LManagedPointer<LCursor> cursor = args.createCursor ();

  cursor->setToFirst();
  LManagedPointer<Sequence> validCommand = args.elementAt ( cursor );

  cursor->setToNext();
  LManagedPointer<Sequence> doCommand = args.elementAt ( cursor );

  LVariant rc;

  while ( (bool)validCommand->exec ( context ) )
    {
      rc = doCommand->exec ( context );
    }

  return rc;
}



// ========>>>>>>>>>>>> Sequence GROUP


LInterpreter::Sequence_Group::~Sequence_Group ()
{
}

LVariant LInterpreter::Sequence_Group::exec ( LInterpreter::Context & context )
{
  LVariant ret;

  LManagedPointer<LCursor> cursor = args.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() )
    {
      ret = args.elementAt ( cursor )->exec ( context );
    }

  return ret;
}

// ========>>>>>>>>>>>> Sequence COMMAND

LInterpreter::Sequence_Command::~Sequence_Command ()
{
}

LInterpreter::Sequence_Command::Sequence_Command ( LString const & _name ) :
  name ( _name )
{
}

LVariant LInterpreter::Sequence_Command::exec ( LInterpreter::Context & context )
{
  LHeap<LVariant> _args;

  LManagedPointer<LCursor> cursor = args.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() )
    {
      _args.add ( args.elementAt ( cursor )->exec( context ) );
    }

  LTable<LVariant> argTable ( _args.entries() );
  int i = 0;

  cursor = _args.createCursor ();
  for ( cursor->setToFirst(); cursor->isValid(); cursor->setToNext() )
    {
      argTable[i] = _args.elementAt ( cursor );
      i++;
    }

  //  LManagedPointer<Command> command = context.getInterpreter().getCommand ( name );
  LManagedPointer<Command> command = context.getCommand ( name );
  if ( command.isValid() )
    {
      LTRACE ( "LInterpreter::Sequence_Command", "Executing command: " + name + " with " + LString(_args.entries() ) + " parameter(s)" );
      return command->exec ( context, argTable );
    } else {
      throw Exception ( Exception::CMD_INVALID, name );
    }
  return "";
}

LInterpreter::Command::Command ( LString const & _name ) :
  name ( _name )
{
}

LInterpreter::Command::~Command ()
{
}

// LInterpreter::Command

/*!
 * \return the name of the command
 *
 * get the name of the command
 */
LString LInterpreter::Command::getName () const
{
  return name;
}

bool LInterpreter::Command::isValid () const
{
  return true;
}

// LInterpreter::ScriptCommand

LInterpreter::ScriptCommand::ScriptCommand ( LString const & _name, LManagedPointer<Sequence>_seq ) :
  Command ( _name ),
  sequence ( _seq )
{
}

LInterpreter::ScriptCommand::~ScriptCommand ()
{
}

LVariant LInterpreter::ScriptCommand::exec ( Context & context, LTable<LVariant> const & _args )
{
  context.pushStack ();
  LHash<LString,LVariant> args;
  for ( unsigned long i = 0; i<_args.getSize(); i++ )
    {
      args.set ( i, _args[i] );
    }
  context.getLocal ( "local_args" ) = args;
  LVariant ret = sequence->exec ( context );
  context.popStack ();
  return ret;
}

bool LInterpreter::ScriptCommand::isValid () const
{
  if ( !sequence.isValid() )
    return false;

  return true;
}

LInterpreter::~LInterpreter ()
{
}


/*!
 * \param command the command to register
 *
 * Registers a new command with the instance. The name will be provided by the Command instance.
 */

void LInterpreter::registerCommand ( LManagedPointer<LInterpreter::Command> command )
{
  if ( !command->isValid () )
    return;

  LTRACE ( "LInterpreter", "registered command " + command->getName() );
  commandHash.set ( command->getName(), command );
}

/*!
 * \param name the name of a registered command
 *
 * Removes the command from the LInterpreter instance.
 *
 * If the command is not registered the call will be ignored.
 */

void LInterpreter::unregisterCommand ( LString const & name )
{
  commandHash.unset ( name );
  LTRACE ( "LInterpreter", "unregistered command " + name );
}

/*!
 * \return the reference to the command instance
 * \param name the name of the command to find
 */

LManagedPointer<LInterpreter::Command> LInterpreter::getCommand ( LString const & name ) const
{
  return commandHash.get ( name, NULL );
}

/*!
 * \return the result of the sequence
 * \param sequence the sequence to execute
 *
 * This method starts the execution of a sequence. The command execution will be initiated
 * by a derieved class of Sequence. Commands are looked up from the LInterpreter instance
 * and will be executed by the Command derieved class.
 */
LVariant LInterpreter::run ( LManagedPointer<Sequence> sequence )
{
  LTRACE_FUNC("LInterpreter", "run");
  Context context ( *this );
  return context.run ( sequence );
}

/*!
 * \return the context hash requested
 * \param the name of the context hash to find
 *
 * A context hash is a custom variable storage which may be shared between commands
 */
/*
LHash<LString,LString> & LInterpreter::getContextHash ( LString const & name )
{
  if ( !contextHash.isSet ( name ) )
    contextHash.set ( name, LHash<LString,LString> () );

  return contextHash.get ( name );
}
*/

// ============>>>>>>>>>> CONTEXT

LInterpreter::Context::Context ( LInterpreter & _inter ) :
  inter ( _inter ),
  localHash ( 32 )
{
}

void LInterpreter::Context::registerCommand ( LManagedPointer<Command> command )
{
  if ( !command->isValid () )
    return;

  LTRACE ( "LInterpreter::Context", "registered command " + command->getName() );
  commandHash.set ( command->getName(), command );
}

void LInterpreter::Context::unregisterCommand ( LString const & name )
{
  commandHash.unset ( name );
  LTRACE ( "LInterpreter", "unregistered command " + name );
}

LManagedPointer<LInterpreter::Command> LInterpreter::Context::getCommand ( LString const & name ) const
{  
  LManagedPointer<Command> cmd = commandHash.get ( name, NULL );
  if ( !cmd.isValid() )
    return getInterpreter().getCommand ( name );
  else
    return cmd;
}

LInterpreter::Context::~Context ()
{
}

LVariant LInterpreter::Context::run ( LManagedPointer<Sequence> sequence )
{
  LTRACE_FUNC("LInterpreter::Context","run");

  pushStack ();
  LVariant ret =  sequence->exec ( *this );
  popStack ();

  return ret;
}

void LInterpreter::Context::clear ( void )
{
  globalHash.clear ();
  localHash.clear ();
}

LInterpreter & LInterpreter::Context::getInterpreter ()
{
  return inter;
}

LInterpreter const & LInterpreter::Context::getInterpreter () const
{
  return inter;
}

LHash<LString,LVariant> & LInterpreter::Context::getGlobal ( LString const & name )
{
  if ( !globalHash.isSet ( name ) )
    globalHash.set ( name, LHash<LString,LVariant> () );

  return globalHash.get ( name );
}

void LInterpreter::Context::pushStack ()
{
  localHash.push ( LHash<LString, LHash<LString,LVariant> > () );
}

void LInterpreter::Context::popStack ()
{
  localHash.pop ();
}

LHash<LString,LVariant> & LInterpreter::Context::getLocal ( LString const & name )
{
  LHash<LString, LHash < LString, LVariant > > & hash = localHash.peek ();
  if ( !hash.isSet ( name ) )
    hash.set ( name, LHash<LString,LVariant> () );

  return hash.get ( name );
}


