#include <lcl.h>
#include <lcl/lloghandler.h>

#ifndef PURE_WIN32
#include <syslog.h>
#endif

#include <lcl/ldatetime.h>

LLogHandler::LLogHandler () :
  name ( "unassigned" ),
  active ( true )
{
}

LString LLogHandler::getName () const
{
  return name;
}

bool LLogHandler::getActive () const
{
  return active;
}

void LLogHandler::setName ( LString const & _name )
{
  name = _name;
}

void LLogHandler::setActive ( bool _active )
{
  active = _active;
}

#ifndef PURE_WIN32
bool LLogHandler_Syslog::initstate = false;

LLogHandler_Syslog::LLogHandler_Syslog ( ) :
  LLogHandler ()
{
  initstate = false;
  setName ( "Syslog" );
}

LLogHandler_Syslog::~LLogHandler_Syslog ()
{
  closelog ();
}

void LLogHandler_Syslog::logEvent ( LLogSystem & system, LLogEvent & event )
{
  if ( !initstate ) {
    openlog ( "LCL->" + system.getModule (), LOG_PID, LOG_USER );
    initstate = true;
  }

  int pri;

  switch ( event.pri ) {

  case LLogSystem::INFO:
    pri = LOG_INFO;
    break;
  case LLogSystem::WARNING:
    pri = LOG_WARNING;
    break;
  case LLogSystem::ERR:
    pri = LOG_ERR;
    break;
  case LLogSystem::FATAL:
    pri = LOG_EMERG;
    break;

  default:
    pri = -1;
    break;

  }

  if ( pri == -1 )
    return;

  syslog ( pri, event.message );
}
#endif

LLogHandler_Console::LLogHandler_Console ( long _level, LOStream & _os ) :
  out ( _os ),
  level ( _level )
{
  setName ( "Console" );
}

LLogHandler_Console::~LLogHandler_Console ()
{
}

void LLogHandler_Console::logEvent ( LLogSystem & system, LLogEvent & event )
{
  if ( event.pri>level )
    return;

  LString buffer;
  buffer = LDateTime(event.time).asString();
  switch ( event.pri ) {

  case LLogSystem::INFO:
    buffer += " [INFO ] ";
    break;
  case LLogSystem::WARNING:
    buffer += " [WARN ] ";
    break;
  case LLogSystem::ERR:
    buffer += " [ERROR] ";
    break;
  case LLogSystem::FATAL:
    buffer += " [FATAL] ";
    break;
  case LLogSystem::BULK:
    buffer += " [BULK ] ";
    break;
  default:
    buffer += " [NONE ] ";
    break;
  }
  buffer += system.getModule() + "[" + LString(event.pid) + "] : ";
  buffer += event.message + "\n";
  out << buffer;
  out << flush;
}

#ifdef _WIN32

#include <winbase.h>

LLogHandler_WinNT::LLogHandler_WinNT ( LString const & _server )
{
  setName ( "WinNT" );
  server = _server;
  handle = 0;
}

void LLogHandler_WinNT::logEvent ( LLogSystem & system, LLogEvent & event )
{
  if ( !handle ) {
    //    handle = OpenEventLog ( server.length() ? ((const char*)server) : NULL, "LCL->" + system.getModule() );
    lout << "Open log" << endl;
    handle = RegisterEventSource ( server.length() ? ((const char*)server) : NULL, "LCL->" + system.getModule() );
  }

  int pri;

  switch ( event.pri ) {

  case LLogSystem::INFO:
    pri = EVENTLOG_INFORMATION_TYPE;
    break;
  case LLogSystem::WARNING:
    pri = EVENTLOG_WARNING_TYPE;
    break;
  case LLogSystem::ERR:
    pri = EVENTLOG_ERROR_TYPE;
    break;
  case LLogSystem::FATAL:
    pri = EVENTLOG_ERROR_TYPE;
    break;

  default:
    pri = -1;
    break;

  }

  if ( pri == -1 )
    return;

  char * str_list [2] = { event.message, NULL };

  if ( ! ReportEvent ( handle, pri, 0, 0, NULL, 1, 0, (char const **)str_list, NULL ) ) {
    lout << "error" << endl;
  }

}

LLogHandler_WinNT::~LLogHandler_WinNT ()
{
  if ( handle )
    //CloseEventLog ( handle );
    DeregisterEventSource ( handle );
}

#endif
