
#include <lcl/llogsystem.h>

#include <lcl/lfile.h>
#include <lcl/lutils.h>
#include <lcl/lstreamproxy.h>
#include <lcl/lfstream.h>

#include <time.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


LCL_EXPORT LLogSystem * LLogSystem::theInstance = NULL;

LLogSystem * LLogSystem::getInstance ( )
{
  if ( !theInstance )
    theInstance = new LLogSystem;

  return theInstance;
}

LLogSystem::LLogSystem ()
{
}

void LLogSystem::setModule ( LString const & _module )
{
  module = _module;
}

LString LLogSystem::getModule ( ) const
{
  return module;
}

void LLogSystem::addLogHandler ( LLogHandler * _handler )
{
  LAutoLock _lock ( syncLock );

  handlerSet.add ( _handler );
}

void LLogSystem::removeLogHandler ( LLogHandler * _handler )
{
  LAutoLock _lock ( syncLock );

  handlerSet.removeElement ( _handler );
}

/*!
 * Log an event to the LCL logging system. Refer to the macro LLOG().
 */
void LLogSystem::log ( LogPriority pri, LString const & _event )
{
  LAutoLock _lock ( syncLock );

  LLogEvent event;
  event.pri = pri;
  event.message = _event;
  event.module = module;
  event.pid = LGetProcessID ();
  event.time = time(0);

  LCursor * cursor = handlerSet.createCursor ();
  for ( cursor->setToFirst (); cursor->isValid(); cursor->setToNext() ) {
    if ( handlerSet.elementAt ( cursor )->getActive() )
      handlerSet.elementAt ( cursor )->logEvent ( *this, event );
  }
  delete cursor;
}

void LLogSystem::dump ()
{
  LLogHandler * handler;

  lout << "-------------- DUMP OF LOGSYSTEM ----------------" << endl;

  LCursor * cursor = handlerSet.createCursor ();
  for ( cursor->setToFirst (); cursor->isValid(); cursor->setToNext() ) {
    handler = handlerSet.elementAt ( cursor );
    lout << handler->getName() << " -> " << ( handler->getActive() ? "active" : "inactive" ) << endl;
  }

  delete cursor;
  lout << "----------------- END OF DUMP -------------------" << endl;
  lout << flush;
}

void LLogSystem::createLogHandler ( LString const & _cmd )
{
  LString cmd = _cmd;
  cmd.setDelims ( "?" );

  LString handler = cmd.word(0).toUpper();
  LString params = cmd.word(1);
  LHash<LString,LString> paramHash;
  params.setDelims ( "+" );

  LString kv;

  for ( int i = 0; i<params.numWords(); i++ ) {
    kv = params.word(i);
    kv.setDelims ( "=" );
    paramHash.set ( kv.word(0).toUpper(), kv.word(1) );
  }

  #ifdef _WIN32
  if ( handler == "WINNT" ) {
    LLogSystem::getInstance () -> addLogHandler ( new LLogHandler_WinNT );
    return;
  }
  #endif
  #ifndef PURE_WIN32
  if ( handler == "SYSLOG" ) {
    LLogSystem::getInstance () -> addLogHandler ( new LLogHandler_Syslog );
    return;
  }
  #endif

  if ( handler == "CONSOLE" ) {

    long level;
    LString lev = paramHash.get("LEVEL","").toUpper();
    if ( lev == "BULK" ) {
      level = LLogSystem::BULK;
    } else if ( lev == "INFO" ) {
      level = LLogSystem::INFO;
    } else if ( lev == "WARNING" ) {
      level = LLogSystem::WARNING;
    } else if ( lev == "ERROR" ) {
      level = LLogSystem::ERR;
    } else if ( lev == "FATAL" ) {
      level = LLogSystem::FATAL;
    } else if ( lev == "" ) {
      level = LLogSystem::BULK;
    } else {
      lout << "Unkown log level '" << lev << "'" << endl;
    }

    LOStream * os = &lout;
    if ( paramHash.isSet ("FILE") ) {
      LFile logFile ( paramHash.get("FILE").replaceWithEnvs(), LFDManager::create | LFDManager::append | LFDManager::write, LFDManager::all_r | LFDManager::user_w );
      if ( logFile.isValid() )
	os = new LFOStream ( logFile );
      else
	lout << "WARNING: Could not open file: " << paramHash.get("FILE").replaceWithEnvs() << " for logging!" << endl << flush;

    }
    LLogSystem::getInstance () -> addLogHandler ( new LLogHandler_Console ( level, *os ) );

  } else {
    lout << "Log Handler '" << handler << "' not found!" << endl;
  }// LOG HANDLERS
}

