#include <lcl/lsem.h>

#include <lcl/ltrace.h>
#include <lcl/lutils.h>

#ifdef _WIN32

#define SEM_MAX 0x7FFFFFFF-1

LSemaphore::LSemaphore ( unsigned long count )
{
  //sem = CreateSemaphore ( NULL, 0, 2147483648, NULL);
  sem = CreateSemaphore ( NULL, count, SEM_MAX, NULL);
  if ( sem == NULL )
    {
      LTRACE ( "LSemaphore", "Failed to create semaphore" );
      LTRACE_PERROR();
    }
}

LSemaphore::~LSemaphore ()
{
  CloseHandle ( sem );
}

void LSemaphore::post ( long value )
{
  //LONG val;
  //ReleaseSemaphore ( sem, 1, &val );
  //cout << "Sem Post: " << val+1 << "\n"; cout.flush ();
  
  if ( !ReleaseSemaphore ( sem, value, NULL ) )
    {
      LTRACE ( "LSemaphore", "Failed to release semaphore" );
      LTRACE_PERROR();
    }  
}

bool LSemaphore::wait ( int to )
{
  DWORD rc;
  if ( to == INFINITE )   
    rc = WaitForSingleObject ( sem, INFINITE );
  else
    rc = WaitForSingleObject ( sem, to*1000 );

  return rc != WAIT_TIMEOUT;
}

bool LSemaphore::trywait ()
{
   return WaitForSingleObject ( sem, 0 ) == WAIT_OBJECT_0;
}

#else

#include <sched.h>
#include <time.h>

LSemaphore::LSemaphore ( unsigned long count )
{
  memset ( &sem, 0, sizeof ( sem ) );
  sem_init ( &sem, 0, count );
}

LSemaphore::~LSemaphore ()
{
  sem_destroy ( &sem );
}

void LSemaphore::post ( long value )
{
  sem_post ( &sem );
}

bool LSemaphore::trywait ( )
{
  return sem_trywait ( &sem ) != -1;
}


bool LSemaphore::wait ( int to )
{
  time_t starttime, currenttime;
  starttime = currenttime = time(0);

  if ( to==0 ) {

    return sem_wait ( &sem ) != -1;

  } else {

    while ( difftime ( currenttime, starttime) < to ) {
      currenttime = time (0);
      if ( trywait () )
	return true;
      sched_yield();      
    }
    return false;
  }
}


#endif


LSemaphoreArray::LSemaphoreArray ( long _size )
{
  size = _size;
  sems = new LSemaphore* [size];

  memset ( sems, 0, sizeof(LSemaphore*) * size );
}

LSemaphoreArray::~LSemaphoreArray ( )
{
  //delete [size] sems;
  delete sems;
}

#ifdef _WIN32

void LSemaphoreArray::waitAll ()
{
  HANDLE * handles = new HANDLE [size];
  long count = 0;
  DWORD rc;

  for ( int i=0; i<size; i++) {
    if ( sems[i] ) {
      handles[count] = sems[i]->sem;
      count++;
    }
  }
  
  rc = WaitForMultipleObjects ( count, handles, true, 10 );

  delete [size] handles;
}

long LSemaphoreArray::waitForNext (void)
{
  DWORD ulong;
  HANDLE * handles = new HANDLE [size];
  long count = 0;

  for ( int i=0; i<size; i++) {
    if ( sems[i] ) {
      handles[count] = sems[i]->sem;
      count++;
    }
  }
  
  ulong = WaitForMultipleObjects ( count, handles, false, INFINITE );
  
  delete [size] handles;
  
  if (ulong == WAIT_FAILED)
    return -1;
  
  return ulong - WAIT_OBJECT_0;
}

#endif // _WIN32

void LSemaphoreArray::set ( long pos, LSemaphore * sem )
{
  if ( pos >= size )
    return;
  
  sems [ pos ] = sem;
}

LSemaphore * LSemaphoreArray::get ( long pos )
{
  if ( pos >= size )
    return NULL;
  return sems[pos];
}
