#include <lcl/lsimplelist.h>
#include <iostream.h>
#include <malloc.h>


/*!
 * Constructs a empty list
 */
LSimpleList::LSimpleList ()
{
  head = tail = NULL;
  locked = false;
}

/*!
 * Copy Constructor
 *
 * \param p1 The list to copy
 * \post A new list containing all elements of the specified list
 */
LSimpleList::LSimpleList ( LSimpleList const & p1 )
{
  head = tail = NULL;
  locked = p1.locked;

  Container * node = p1.head;
  Container * temp;

  while ( node ) {
    temp = new Container ( *node );
    append ( temp );
    node = node->next;
  }
}

LSimpleList::~LSimpleList ()
{  
  clear ();
}

/*!
 * \param ptr the value to store
 * \post contains a new element appended at the end of the list
 *
 * This method adds a new value at the end of the list. The new value
 * will be encapsulated in a LSimpleList::Container which is automatically
 * created
 */
void LSimpleList::append ( void * ptr )
{
  Container * node = new Container;
  node->ptr = ptr;

  if ( tail ) {
    node->prev = tail;
    node->next = NULL;
    tail->next = node;
    tail = node;
  }  else {
    node->prev = NULL;
    node->next = NULL;
    tail = node;
    head = node;
  }

}

/*!
 * Removes a element from the list
 *
 * \param ptr the value to search and remove
 * \post the element will be removed from the list if it exists otherwise nothing will happend.
 * \post If the list is currently locked the element will only be marked as deleted. It must be deleted using LSimpleList::cleanUp
 */
void LSimpleList::remove ( void * ptr )
{
  if ( head == NULL )
    return;

  Container * node = head;
  Container * temp;

  while ( node ) {

    temp = node->next;

    // Check if this is a node to process
    if (node->ptr == ptr) {

      if ( locked ) {
	
	// Set node to zero (will be found as node when remove(NULL) is called
	node->ptr = NULL;

      } else {

	// Remove this node
	
	if ( node->prev ) {
	  node->prev->next = node->next;
	} else {
	  if (node->next)
	    node->next->prev = NULL;
	  head = node->next;
	}
	
	if ( node->next ) {
	  node->next->prev = node->prev;
	} else {
	  if ( node->prev )
	    node->prev->next = NULL;
	  tail = node->prev;	
	}
	
	// Delete Container
	delete node;

      } // Node deleted

    }
      
    node = temp;
  }
}

/*!
 * \post the list is completly empty
 * \warning This method ignores the locked state
 *
 * Clears the whole list
 */

void LSimpleList::clear ()
{
  Container * node, *temp;

  node = head;

  while ( node ) {    
    temp = node->next;
    delete node;
    node = temp;
  }

  head = tail = NULL;  

}

/*!
 * Iterates through all elements of the list
 *
 * \param func the function to call for each item
 */
void LSimpleList::forEach ( void (*func)(void*) )
{
  Container * node;
  
  node = head;

  while (node) {
    if (node->ptr)
      func ( node->ptr );
    node = node->next;
  }

  cleanUp ();
}

/*!
 * \post the list is cleaned from entries marked as deleted
 *
 * When the list is locked and an element is deleted it will only
 * be marked as status deleted. This method cleans up all entires
 * mark as deleted
 */
void LSimpleList::cleanUp ( )
{
  if ( !locked )
    remove(NULL);
}

/*!
 * Lock the list (also see LSimpleList::cleanUp and LSimpleList::remove)
 */

void LSimpleList::lock ()
{
  locked = true;
}

/*!
 * UnLock the list (also see LSimpleList::cleanUp and LSimpleList::remove)
 *
 * \post The list is un-locked and also cleaned up from marked elements
 */

void LSimpleList::unlock ()
{
  locked = false;
  cleanUp();
}
