#include <lcl/lstringtokenizer.h>

#include <lcl/lpointer.h>

LStringTokenizer::Token::Token () :
  str ( NULL )
{
}

LStringTokenizer::Token::Token ( LString * _str, long _num, long _start, long _len ) :
  num ( _num ),
  start ( _start ),
  len ( _len ),
  str ( _str )
{
}

LStringTokenizer::Token::Token ( Token const & p1 ) :
  num ( p1.num ),
  start ( p1.start ),
  len ( p1.len ),
  str ( p1.str )
{
}

bool LStringTokenizer::Token::isValid () const
{
  return str != NULL;
}

LStringTokenizer::LStringTokenizer ( LString const & _delims, LString const & _qmarks ) :
  delims ( _delims ),
  qmarks ( _qmarks ),
  words ( 0 )
{  
}

/*!
 * This function will first clear the tokenizer and then tokenize the string passed.
 * It will use the settings which where passed to the constructor.
 *
 * \b Example:
 * \verbatim
 '   This is the sentence   to tokenize "quickly and smooth!"   '
     ---- -- --- --------   -- --------  -------------------
      #0  #1  #2    #3      #4   #5            #6

 '   The answer "is 42! '
     --- ------  -------
      #0   #1      #2

 'I "   " "  " am "   " bored!!'
  -  ---   --  --  ---  -------
  #0  #1   #2  #3   #4    #5 

\endverbatim
 *
 * \param _str the string to tokenize
 */
void LStringTokenizer::tokenize ( LString const & _str )
{
  clear ();
  str = _str;

  // cleared

  bool lastWS = true;
  bool qmarked = false;
  char c;

  long start = 0, len = 0, num = 0;

  long i, strlen = str.length();
  for ( i = 0; i<strlen; i++ ) {

    c = str.get(i);

    if ( qmarks.contains ( c ) ) {
      qmarked = !qmarked;
      //      cout << ( qmarked ? "(" : ")" );

    } else if ( delims.contains ( c ) && !qmarked ) {
      if ( lastWS ) {
	//	cout << "_";
      } else {
	//	cout << "-";
	//	cout << "Adding token " << num << " len " << len << " start " << start << endl;
	tokenList.add ( Token ( &str, num, start, len ) );
	num++;
	lastWS = true;
      }

    } else {
      if ( lastWS ) {
	lastWS = false;
	start = i;
	len = 0;
      }
      len++;
      //      cout << c;
    }

  }

  if ( !lastWS ) {
    //    cout << "Adding token " << num << " len " << len << " start " << start << endl;
    tokenList.add ( Token ( &str, num, start, len ) );
    num++;
  }
  words = num;
}

/*!
 * Clears all internal structures. Any queries to words or the number of words will fail till the next
 * tokenization.
 */
void LStringTokenizer::clear ()
{
  str = "";
  tokenList.clear ();
  words = 0;  
}

/*!
 * Returns the number of words from the last tokenization.
 * \return the number of words
 */
long LStringTokenizer::numWords () const
{  
  return words;
}

LStringTokenizer::Token LStringTokenizer::getTokenInfo ( long pos ) const
{
  LManagedPointer<LCursor> cursor = tokenList.createCursor ();

  for ( cursor->setToFirst() ; cursor->isValid() ; cursor->setToNext () ) {
    if ( tokenList.elementAt ( cursor ).num == pos )
      return tokenList.elementAt ( cursor );
  }

  return Token();
}

/*!
 *
 * Returns the specified word. The index starts with 0. If the word number is greater than found the
 * number of words found it will return an empty string.
 *
 * \return the specified word
 * \param _num the word number
 */
LString LStringTokenizer::word ( long _num ) const
{
  Token tokenInfo ( getTokenInfo (_num) );

  if ( !tokenInfo.isValid () ) {
    return "";
  }

  /*  if ( tokenInfo.str != &str )
      return "";*/

  return str.subString ( tokenInfo.start, tokenInfo.len );
}

bool LStringTokenizer::containsWord ( LString const & _word, bool caseSens ) const
{  
  long i;
  LString myWord = _word.toUpper();

  for ( i = 0; i<numWords(); i++ ) {
    if ( !caseSens ) {
      if ( myWord == word(i).toUpper() )
	return true;
    } else {
      if ( _word == word(i) )
	return true;
    }
  }
  return false;
}


LString LStringTokenizer::rightOfWord ( long pos ) const
{
  Token tokenInfo ( getTokenInfo ( pos ) );

  if ( !tokenInfo.isValid () ) {
    return "";
  }

  return str.subString ( tokenInfo.start + tokenInfo.len );
}

LHeap<LString> LStringTokenizer::asHeap () const
{
  LHeap<LString> list;
  long i;
  for ( i = 0; i<numWords(); i++ )
    {
      list.add ( word(i) );
    }
  return list;
}
